import logging

from starlette import status
from marshmallow import Schema, fields

from starlette_web.contrib.auth.models import User
from starlette_web.common.http.base_endpoint import BaseHTTPEndpoint


logger = logging.getLogger(__name__)


class ServicesCheckSchema(Schema):
    postgres = fields.Str()


class HealthCheckSchema(Schema):
    services = fields.Nested(ServicesCheckSchema)
    errors = fields.List(fields.Str)


class HealthCheckAPIView(BaseHTTPEndpoint):
    """Allows controlling status of web application (live ASGI and pg connection)"""

    auth_backend = None
    response_schema = HealthCheckSchema

    async def get(self, *_):
        """
        description: Health check of services
        responses:
          200:
            description: Services with status
            content:
              application/json:
                schema: HealthCheckSchema
          503:
            description: Service unavailable
        tags: ["Health check"]
        """
        response_data = {"services": {}, "errors": []}
        result_status = status.HTTP_200_OK

        try:
            await User.async_filter(self.db_session)
        except Exception as error:
            error_msg = f"Couldn't connect to DB: {error.__class__.__name__} '{error}'"
            logger.exception(error_msg)
            response_data["services"]["postgres"] = "down"
            response_data["errors"].append(error_msg)
        else:
            response_data["services"]["postgres"] = "ok"

        services = response_data.get("services").values()

        if "down" in services or response_data.get("errors"):
            response_data["status"] = "down"
            result_status = status.HTTP_503_SERVICE_UNAVAILABLE

        return self._response(
            data=response_data,
            status_code=result_status,
        )



from typing import Callable, Any

import inspect


def get_available_options(_method: Callable, exclude=("self", )):
    full_args_spec = inspect.getfullargspec(_method)
    result = full_args_spec.args + full_args_spec.kwonlyargs
    return [
        arg
        for arg in result
        if arg not in exclude
    ]


def safe_init(_klass, **options) -> Any:
    args = get_available_options(_klass.__init__)
    return _klass(**{
        key: value
        for key, value in options.items()
        if key in args
    })



import time
import os
from pathlib import Path

import pytest

from starlette_web.common.conf import settings
from starlette_web.common.files.storages import FilesystemStorage
from starlette_web.tests.helpers import await_


class TestFileSystemStorage:
    base_dir = settings.FILESTORAGE_DIR

    def test_file_write(self):
        rel_path = "dir1/dir2/file1.txt"

        async def write_task():
            async with FilesystemStorage(BASE_DIR=self.base_dir) as storage:
                async with storage.writer(rel_path, "t") as writer:
                    await writer.write("Test content")

        await_(write_task())
        assert (Path(self.base_dir) / rel_path).is_file()

    def test_file_read(self):
        rel_path = "dir1/dir3/file2.txt"

        async def write_and_read_task():
            async with FilesystemStorage(BASE_DIR=self.base_dir) as storage:
                async with storage.writer(rel_path, "b") as writer:
                    await writer.write(b"Test content")

                async with storage.reader(rel_path, "b") as reader:
                    return await reader.read(4)

        value = await_(write_and_read_task())
        assert (Path(self.base_dir) / rel_path).is_file()
        assert value == b"Test"

    def test_readline(self):
        rel_path = "dir1/dir3/file5.txt"

        with open(Path(self.base_dir) / rel_path, "wb") as file:
            file.write(b"\n".join([
                (b"Line " + str(i).encode())
                for i in range(1000)
            ]))

        async def write_and_read_task():
            _lines = []

            async with FilesystemStorage(BASE_DIR=self.base_dir) as storage:
                async with storage.reader(rel_path, "b") as _reader:
                    async for line in _reader:
                        _lines.append(line)

            return _lines

        lines = await_(write_and_read_task())
        assert lines[0].strip(b"\r\n") == b"Line 0"
        assert len(lines) == 1000

    def test_file_delete(self):
        rel_path = "dir1/dir2/file4.txt"

        async def create_task():
            async with FilesystemStorage(BASE_DIR=self.base_dir) as storage:
                async with storage.writer(rel_path, "b") as writer:
                    await writer.write(b"Test content")

        await_(create_task())
        assert (Path(self.base_dir) / rel_path).is_file()
        assert (Path(self.base_dir) / "dir1").is_dir()

        async def delete_task():
            async with FilesystemStorage(BASE_DIR=self.base_dir) as storage:
                await storage.delete(rel_path)

        await_(delete_task())
        assert not (Path(self.base_dir) / rel_path).is_file()

        async def delete_directory():
            async with FilesystemStorage(BASE_DIR=self.base_dir) as storage:
                for path in (await storage.listdir("dir1/dir2")):
                    await storage.delete("dir1/dir2/" + path)

                await storage.delete("dir1/dir2")

        await_(delete_directory())
        assert not (Path(self.base_dir) / "dir1" / "dir2").is_dir()

    def test_utils_functions(self):
        rel_path = "dir1/dir3/file3.txt"
        current_time = time.time()

        async def utils_task():
            async with FilesystemStorage(BASE_DIR=self.base_dir) as storage:
                async with storage.writer(rel_path, "b") as writer:
                    await writer.write(b"Test content")

                _mtime = await storage.get_mtime(rel_path)
                _exists = await storage.exists(rel_path)
                _listdir = await storage.listdir(os.path.dirname(rel_path))
                _size = await storage.size(rel_path)

                return _mtime, _exists, _listdir, _size

        mtime, exists, listdir, size = await_(utils_task())

        assert abs(current_time - mtime) < 1.0
        assert exists
        assert "file3.txt" in listdir
        assert size == 12

    def test_cant_delete_non_empty_dir(self):
        rel_path = "dir1/dir2/file1.txt"

        async def write_and_delete_task():
            async with FilesystemStorage(BASE_DIR=self.base_dir) as storage:
                async with storage.writer(rel_path, "t") as writer:
                    await writer.write("Test content")

                await storage.delete("dir1/dir2")

        with pytest.raises(OSError):
            await_(write_and_delete_task())



class BaseUserMixin:
    # Contrary to django's AbstractBaseUser, this is not bound to DB,
    # therefore a mixin

    @property
    def is_authenticated(self) -> bool:
        raise NotImplementedError


class AnonymousUser(BaseUserMixin):
    @property
    def is_authenticated(self) -> bool:
        raise False



import uuid

from starlette_web.common.management.base import BaseCommand
from starlette_web.contrib.auth.models import User


class Command(BaseCommand):
    help = "Command to test database connection"

    async def handle(self, **options):
        test_user_password = str(uuid.uuid4())
        email = str(uuid.uuid4()).replace("-", "") + "@test.com"
        password = User.make_password(test_user_password)

        async with self.app.session_maker() as session:
            user = await User.async_create(session, email=email, password=password)

            await User.async_delete(session, {"id": user.id})
            await session.commit()



import logging
import sys
from typing import Type, Union, Iterable, ClassVar, Optional, Mapping, List

from marshmallow import Schema, ValidationError
from sqlalchemy.ext.asyncio import AsyncSession
from starlette import status
from starlette.background import BackgroundTasks
from starlette.exceptions import HTTPException
from starlette.endpoints import HTTPEndpoint
from starlette.requests import Request
from webargs_starlette import WebargsHTTPException, StarletteParser

from starlette_web.common.app import WebApp
from starlette_web.common.authorization.backends import (
    BaseAuthenticationBackend,
    NoAuthenticationBackend,
)
from starlette_web.common.authorization.permissions import PermissionType
from starlette_web.common.authorization.base_user import AnonymousUser
from starlette_web.common.conf import settings
from starlette_web.common.http.exceptions import (
    UnexpectedError,
    BaseApplicationError,
    InvalidParameterError,
    PermissionDeniedError,
)
from starlette_web.common.http.renderers import BaseRenderer
from starlette_web.common.database import DBModel
from starlette_web.common.utils import import_string


logger = logging.getLogger(__name__)


class BaseHTTPEndpoint(HTTPEndpoint):
    """
    Base View witch used as a base class for every API's endpoints
    """

    app = None
    request: Request
    db_session: AsyncSession
    request_schema: ClassVar[Type[Schema]]
    response_schema: ClassVar[Type[Schema]]
    auth_backend: ClassVar[Type[BaseAuthenticationBackend]] = NoAuthenticationBackend
    permission_classes: ClassVar[List[PermissionType]] = []
    request_parser: ClassVar[Type[StarletteParser]] = \
        import_string(settings.DEFAULT_REQUEST_PARSER)
    response_renderer: ClassVar[Type[BaseRenderer]] = \
        import_string(settings.DEFAULT_RESPONSE_RENDERER)

    async def dispatch(self) -> None:
        """
        This method is called in every request.
        So, we can use this one for custom authentication and exception handling
        """

        self.request = Request(self.scope, receive=self.receive)
        self.app: WebApp = self.scope.get("app")

        handler_name = "get" if self.request.method == "HEAD" else self.request.method.lower()
        handler = getattr(self, handler_name, self.method_not_allowed)

        try:
            session_maker = self.app.session_maker()
            session = await session_maker.__aenter__()
        except Exception as err:
            msg_template = "Unexpected error handled: %r"
            logger.exception(msg_template, err)
            raise UnexpectedError(msg_template % (err,))

        try:
            self.request.state.db_session = session
            self.db_session = session

            await self._authenticate()
            await self._check_permissions()

            response = await handler(self.request)  # noqa
            await session.commit()

        except (BaseApplicationError, WebargsHTTPException, HTTPException) as err:
            await session.rollback()
            raise err

        except Exception as err:
            await session.rollback()
            msg_template = "Unexpected error handled: %r"
            logger.exception(msg_template, err)
            raise UnexpectedError(msg_template % (err,))

        finally:
            await session_maker.__aexit__(*sys.exc_info())
            self.request.state.db_session = None

        await response(self.scope, self.receive, self.send)

    async def _authenticate(self):
        if self.auth_backend:
            backend = self.auth_backend(self.request, self.scope)
            user = await backend.authenticate()
            self.scope["user"] = user
        else:
            self.scope["user"] = AnonymousUser()

    async def _check_permissions(self):
        for permission_class in self.permission_classes:
            try:
                has_permission = await permission_class().has_permission(self.request, self.scope)
                if not has_permission:
                    raise PermissionDeniedError
            # Exception may be raised inside permission_class, to pass additional details
            except PermissionDeniedError as exc:
                raise exc
            except Exception as exc:
                raise PermissionDeniedError from exc

    async def _validate(
        self, request, schema: Type[Schema] = None, partial_: bool = False, location: str = None
    ) -> Optional[Mapping]:
        """Simple validation, based on marshmallow's schemas"""

        schema_class = schema or self.request_schema
        schema_kwargs = {}
        if partial_:
            schema_kwargs["partial"] = [field for field in schema_class().fields]

        schema_obj, cleaned_data = schema_class(**schema_kwargs), {}
        try:
            cleaned_data = await self.request_parser().parse(schema_obj, request, location=location)
            if hasattr(schema_obj, "is_valid") and callable(schema_obj.is_valid):
                schema_obj.is_valid(cleaned_data)

        except ValidationError as e:
            # TODO: check that details is str / flatten
            raise InvalidParameterError(details=e.data)

        return cleaned_data

    def _response(
        self,
        data: Union[DBModel, Iterable[DBModel], dict] = None,
        status_code: int = status.HTTP_200_OK,
        headers: Mapping[str, str] = None,
        background: Optional[BackgroundTasks] = None,
    ) -> BaseRenderer:
        """
        A shorthand for response_renderer plus serializing data and passing text status.
        To be used primarily with JSONRenderer and such.
        """
        if (data is not None) and self.response_schema:
            schema_kwargs = {}
            if isinstance(data, Iterable) and not isinstance(data, dict):
                schema_kwargs["many"] = True

            payload = self.response_schema(**schema_kwargs).dump(data)
        else:
            payload = data

        return self.response_renderer(
            payload,
            status_code=status_code,
            headers=headers,
            background=background,
        )



import os
import sys

from starlette_web.common.app import get_app
from starlette_web.common.management.base import fetch_command_by_name, CommandError


if __name__ == "__main__":
    settings_module = "core.settings"

    sys_argv = list(sys.argv).copy()
    for arg in sys_argv.copy():
        if arg.startswith("--settings="):
            settings_module = arg[11:]
            sys_argv.remove(arg)

    os.environ.setdefault("STARLETTE_SETTINGS_MODULE", settings_module)

    if len(sys_argv) < 2:
        raise CommandError(
            'Missing command name. Correct syntax is: "python command.py command_name ..."'
        )

    from starlette_web.common.conf import settings

    command = fetch_command_by_name(sys_argv[1])
    app = get_app(use_pool=settings.DB_USE_CONNECTION_POOL_FOR_MANAGEMENT_COMMANDS)
    command(app).run_from_command_line(sys_argv)



import json
from typing import Union, Optional

from httpx import Response
from starlette.testclient import TestClient

from starlette_web.common.database.model_base import ModelBase
from starlette_web.contrib.auth.views import JWTSessionMixin
from starlette_web.contrib.auth.models import UserSession


class BaseTestCase:
    @staticmethod
    def assert_called_with(mock_callable, *args, **kwargs):
        """Check mock object (callable) on call action with provided `args`, `kwargs`"""

        assert mock_callable.called
        mock_call_args = mock_callable.call_args_list[-1]
        if args:
            assert mock_call_args.args == args
        for key, value in kwargs.items():
            assert key in mock_call_args.kwargs, mock_call_args.kwargs
            assert mock_call_args.kwargs[key] == value


class BaseTestAPIView(BaseTestCase):
    url: str = NotImplemented
    default_fail_status_code = 500

    @staticmethod
    def assert_ok_response(response: Response, status_code: int = 200) -> Union[dict, list]:
        assert response.status_code == status_code

        try:
            response_data = response.json()
        except Exception:
            raise AssertionError(response.text)

        assert "error" not in response_data, response_data
        return response_data

    def assert_fail_response(
        self, response: Response, status_code: int = None,
    ) -> Union[dict, list]:

        response_data = response.json()
        assert response.status_code == (status_code or self.default_fail_status_code)
        assert "error" in response_data, response_data
        assert "details" in response_data, response_data
        return response_data

    @staticmethod
    def assert_bad_request(response: Response, error_details: dict):
        response_data = response.json()
        assert response.status_code == 400
        assert "error" in response_data, response_data
        assert "details" in response_data, response_data
        assert response_data["error"] in "Requested data is not valid."

        for error_field, error_value in error_details.items():
            assert error_field in response_data["details"]
            assert error_value in response_data["details"][error_field]

    @staticmethod
    def assert_not_found(response: Response, instance: ModelBase):
        assert response.status_code == 404
        response_data = response.json()
        assert response_data == {
            "error": "Requested object not found.",
            "details": (
                f"{instance.__class__.__name__} #{instance.id} "
                f"does not exist or belongs to another user"
            ),
        }

    def assert_unauth(self, response: Response):
        response_data = self.assert_fail_response(response, status_code=401)
        assert response_data == {
            "error": "Authentication is required.",
            "details": "Invalid token header. No credentials provided.",
        }

    def assert_auth_invalid(
        self,
        response_data: Union[Response, dict],
        details: Optional[str],
        message: Optional[str] = None,
    ):
        if isinstance(response_data, Response):
            response_data = self.assert_fail_response(response_data, status_code=401)

        if message is None:
            message = "Authentication credentials are invalid."

        assert response_data == {
            "error": message,
            "details": details,
        }


class BaseTestWSAPI(BaseTestCase):
    url: str = NotImplemented

    @staticmethod
    def _get_headers(user_session: UserSession) -> dict:
        token_col = JWTSessionMixin._get_tokens(user_session.user, user_session.public_id)  # noqa
        return {"Authorization": f"Bearer {token_col.access_token}"}

    def _ws_request(
        self, client: TestClient, user_session: UserSession, data: dict | None = None
    ) -> dict | list:
        data = {"headers": self._get_headers(user_session)} | (data or {})
        with client.websocket_connect(self.url) as websocket:
            websocket.send_json(data)
            response_data = json.loads(websocket.receive_text())

        return response_data



import logging

from starlette_web.common.management.base import BaseCommand, CommandParser
from starlette_web.contrib.scheduler.backends import get_periodic_scheduler_backend_class


logger = logging.getLogger("starlette_web.contrib.scheduler")


class Command(BaseCommand):
    help = "Manage periodic jobs with OS-wide task scheduler"

    def add_arguments(self, parser: CommandParser):
        parser.add_argument(
            "cmd", choices=["add", "show", "remove", "run"], type=str,
        )
        parser.add_argument("--jobhash", type=str, required=False)

    async def handle(self, **options):
        scheduler_class = get_periodic_scheduler_backend_class()

        if options["cmd"] == "add":
            with scheduler_class(needs_read=True, needs_write=True) as scheduler:
                scheduler.update_jobs()

        elif options["cmd"] == "remove":
            with scheduler_class(needs_read=True, needs_write=True) as scheduler:
                scheduler.remove_jobs()

        elif options["cmd"] == "show":
            with scheduler_class(needs_read=True, needs_write=False) as scheduler:
                scheduler.show_jobs()

        elif options["cmd"] == "run":
            with scheduler_class(needs_read=False, needs_write=False) as scheduler:
                await scheduler.run_job(options["jobhash"])

        else:
            logger.warning(self.help)



import time

import anyio
import pytest

from starlette_web.common.caches.base import BaseCache
from starlette_web.tests.helpers import await_
from starlette_web.common.caches.base import CacheLockError


class BaseCacheTester:
    def _run_base_cache_test(self, cache: BaseCache):
        test_key = "b6e58cf5-4b83-400c-8c11-1dee87174200"
        await_(cache.async_delete(test_key))

        value = await_(cache.async_get(test_key))
        assert value is None

        value_obj = {
            "item_1": None,
            "item_2": True,
            "item_3": 432534523452345234523452435,
            "item_4": ["str", 12, None, True, {}],
            "item_5": {
                "item_5_1": False,
                "item_5_2": 34,
            },
        }

        await_(cache.async_set(test_key, value_obj, 0.2))
        value = await_(cache.async_get(test_key))
        assert value == value_obj
        assert await_(cache.async_has_key(test_key))

        time.sleep(0.3)
        value = await_(cache.async_get(test_key))
        assert value is None

    def _run_cache_many_ops_test(self, cache: BaseCache):
        test_key_1 = "2a8006ac-e296-486e-8044-b27517aaaaaa"
        test_key_2 = "2a8006ac-e296-486e-8044-b27517bbbbbb"

        await_(cache.async_delete_many([test_key_1, test_key_2]))

        keys = await_(cache.async_keys("2a8006ac-e296-486e-8044-b27517*"))
        assert len(keys) == 0

        await_(
            cache.async_set_many(
                {
                    test_key_1: 10,
                    test_key_2: 42,
                },
                timeout=0.2,
            )
        )

        keys = await_(cache.async_keys("2a8006ac-e296-486e-8044-b27517*"))
        assert len(keys) == 2

        values = await_(cache.async_get_many(keys))
        assert values == {
            test_key_1: 10,
            test_key_2: 42,
        }

        await_(cache.async_set(test_key_2, 1.0))
        time.sleep(0.3)

        keys = await_(cache.async_keys("2a8006ac-e296-486e-8044-b27517*"))
        assert len(keys) == 1

    def _run_cache_lock_test(self, cache: BaseCache):
        async def lock_checker():
            async with cache.lock("test_lock", timeout=0.5):
                await anyio.sleep(0.3)

        start_time = time.time()
        await_(lock_checker())
        end_time = time.time()
        assert abs(end_time - start_time - 0.3) < 0.2

    def _run_cache_mutual_lock_test(self, cache: BaseCache):
        async def lock_checker():
            async with cache.lock("test_mutual_lock", timeout=0.2, blocking_timeout=0.1):
                async with cache.lock("test_mutual_lock", timeout=0.2, blocking_timeout=0.1):
                    await anyio.sleep(0.2)

        with pytest.raises(CacheLockError):
            await_(lock_checker())

    def _run_locks_timeouts_test(self, cache: BaseCache):
        timeout = 0.1
        number_of_tests = 4
        sleep_time = 0.2

        async def task_with_lock():
            async with cache.lock("test_lock_timeout", blocking_timeout=2.0, timeout=timeout):
                await anyio.sleep(sleep_time)

        start_time = time.time()

        async def gather_coroutines():
            async with anyio.create_task_group() as nursery:
                for _ in range(number_of_tests):
                    nursery.start_soon(task_with_lock)

        await_(gather_coroutines())
        end_time = time.time()
        run_time = end_time - start_time

        expected_runtime = (sleep_time + (number_of_tests - 1) * timeout)
        assert abs(run_time - expected_runtime) < 0.2

    def _run_base_lock_cancellation(self, cache: BaseCache):
        move_on_after = 0.2

        async def locked_task():
            with anyio.move_on_after(move_on_after):
                async with cache.lock(
                    "test_cancel_lock",
                    blocking_timeout=3.0,
                    timeout=3.0,
                ):
                    await anyio.sleep(3.0)

        start_time = time.time()
        with pytest.raises(CacheLockError):
            await_(locked_task())
        end_time = time.time()
        run_time = end_time - start_time
        assert abs(run_time - move_on_after) < 0.1



import logging
from typing import Dict, Any, Optional

import anyio
from marshmallow import Schema, fields
from marshmallow.validate import OneOf
from starlette.types import Scope, Receive, Send
from starlette.websockets import WebSocket, WebSocketDisconnect

from starlette_web.common.authorization.permissions import IsAuthenticatedPermission
from starlette_web.common.conf import settings
from starlette_web.common.caches import caches
from starlette_web.common.channels.base import Channel
from starlette_web.common.ws.base_endpoint import BaseWSEndpoint
from starlette_web.common.utils.crypto import get_random_string
from starlette_web.contrib.auth.backend import JWTAuthenticationBackend
from starlette_web.contrib.redis.channel_layers import RedisPubSubChannelLayer


logger = logging.getLogger("starlette_web.tests")
locmem_cache = caches["locmem"]


class WebsocketRequestSchema(Schema):
    request_type = fields.Str()


class ChatRequestSchema(Schema):
    request_type = fields.Str(
        validate=[
            OneOf(["connect", "publish"]),
        ]
    )
    message = fields.Str(required=False)


class BaseWebsocketTestEndpoint(BaseWSEndpoint):
    request_schema = WebsocketRequestSchema

    async def _background_handler(self, task_id: str, websocket: WebSocket, data: Dict):
        await anyio.sleep(2)
        return data["request_type"]

    async def _register_background_task(self, task_id: str, websocket: WebSocket, data: Dict):
        await locmem_cache.async_set(task_id, 1, timeout=20)

    async def _unregister_background_task(
        self, task_id: str, websocket: WebSocket, task_result: Any
    ):
        await locmem_cache.async_delete(task_id)
        if task_result is not None:
            await locmem_cache.async_set(task_id + "_result", task_result, timeout=20)


class CancellationWebsocketTestEndpoint(BaseWebsocketTestEndpoint):
    async def _background_handler(self, task_id: str, websocket: WebSocket, data: Dict):
        if data["request_type"] == "cancel":
            return

        if data["request_type"] == "fail":
            raise Exception("fail")

        return await super()._background_handler(task_id, websocket, data)

    async def _handle_background_task_exception(
        self, task_id: str, websocket: WebSocket, exc: Exception
    ):
        await locmem_cache.async_set(task_id + "_exception", str(exc), timeout=20)
        raise WebSocketDisconnect(code=1005) from exc


class AuthenticationWebsocketTestEndpoint(BaseWebsocketTestEndpoint):
    auth_backend = JWTAuthenticationBackend
    permission_classes = [IsAuthenticatedPermission]
    EXIT_MAX_DELAY = 5


class FinitePeriodicTaskWebsocketTestEndpoint(BaseWebsocketTestEndpoint):
    EXIT_MAX_DELAY = 5

    async def _background_handler(self, task_id: str, websocket: WebSocket, data: Dict):
        for i in range(4):
            await websocket.send_json({"response": i})
            await anyio.sleep(1)

        return "finished"


class InfinitePeriodicTaskWebsocketTestEndpoint(BaseWebsocketTestEndpoint):
    EXIT_MAX_DELAY = 5

    async def _background_handler(self, task_id: str, websocket: WebSocket, data: Dict):
        prefix = data["request_type"]

        i = 0
        while True:
            await websocket.send_json({"response": prefix + "_" + str(i)})
            await anyio.sleep(1)
            i += 1

    async def _unregister_background_task(
        self, task_id: str, websocket: WebSocket, task_result: Any
    ):
        await locmem_cache.async_delete(task_id)
        # Explicitly set "finished" task_result for tests
        await locmem_cache.async_set(task_id + "_result", "finished", timeout=20)


class ChatWebsocketTestEndpoint(BaseWSEndpoint):
    request_schema = ChatRequestSchema

    def __init__(self, scope: Scope, receive: Receive, send: Send) -> None:
        super().__init__(scope, receive, send)
        self._manager_lock = anyio.Lock()
        self._response_handler_init = False
        self._channels_init = False
        self._channels_wrap: Optional[Channel] = None
        self._channels: Optional[Channel] = None
        self._tasks = set()

    async def _register_background_task(self, task_id: str, websocket: WebSocket, data: Dict):
        async with self._manager_lock:
            if not self._channels_init and data["request_type"] == "connect":
                redis_options = settings.CHANNEL_LAYERS["redispubsub"]["OPTIONS"]
                self._channels_wrap = Channel(RedisPubSubChannelLayer(**redis_options))
                self._channels = await self._channels_wrap.__aenter__()
                self._channels_init = True

            self._tasks.add(task_id)

    async def _background_handler(self, task_id: str, websocket: WebSocket, data: Dict):
        async with self._manager_lock:
            if not self._channels_init:
                logger.debug("No initialized channels detected. Quit handler")
                return

        room = "chatroom"

        if data["request_type"] == "publish":
            await self._channels.publish(room, data["message"])

        elif data["request_type"] == "connect":
            # In test endpoint, _unregister_background_task checks that
            # there are any tasks left, before closing the channel.
            # Since this parent task will close as soon
            # it spawns dialogue task, we need to register the latter.
            dialogue_task_id = get_random_string(50)
            self._tasks.add(dialogue_task_id)
            self.task_group.start_soon(self._run_dialogue, websocket, room, dialogue_task_id)

        else:
            raise WebSocketDisconnect(code=1005, reason="Invalid request type")

    async def _unregister_background_task(
        self, task_id: str, websocket: WebSocket, task_result: Any
    ):
        async with self._manager_lock:
            self._tasks.discard(task_id)

            if self._channels_init and not self._tasks:
                await self._channels_wrap.__aexit__(None, None, None)
                self._channels_init = False

    async def _run_dialogue(self, websocket: WebSocket, room: str, dialogue_task_id: str):
        try:
            async with self._manager_lock:
                if self._response_handler_init:
                    return
                self._response_handler_init = True

            async with self._channels.subscribe(room) as subscriber:
                async for event in subscriber:
                    await websocket.send_json(event.message)
        finally:
            self._tasks.discard(dialogue_task_id)
            async with self._manager_lock:
                self._response_handler_init = False



import argparse
import datetime
from decimal import Decimal
import uuid
from typing import Callable, Union, Optional, Type


Number = Union[float, int, Decimal]
OptionalNum = Optional[Number]


def arg_uuid(arg_value: str) -> uuid.UUID:
    try:
        return uuid.UUID(arg_value)
    except (ValueError, TypeError):
        raise argparse.ArgumentTypeError("Invalid UUID value")


def arg_date(arg_value: str) -> datetime.date:
    try:
        return datetime.date.fromisoformat(arg_value)
    except ValueError:
        raise argparse.ArgumentTypeError("Invalid ISO date value")


def arg_datetime(arg_value: str) -> datetime.datetime:
    try:
        return datetime.datetime.fromisoformat(arg_value)
    except ValueError:
        raise argparse.ArgumentTypeError("Invalid ISO datetime value")


def arg_decimal(arg_value: str) -> Decimal:
    try:
        return Decimal(arg_value)
    except ValueError:
        raise argparse.ArgumentTypeError("Invalid decimal value")


def arg_range(
    _min=None,
    _max=None,
    _type: Type = int,
) -> Callable[[str], Number]:
    def validator(arg_value: str) -> Number:
        try:
            if _type == int:
                value = int(arg_value)
            elif _type == float:
                value = float(arg_value)
            elif _type == Decimal:
                value = Decimal(arg_value)
            else:
                raise argparse.ArgumentTypeError("Unsupported range type")
        except (ValueError, TypeError):
            raise argparse.ArgumentTypeError(
                f"String {arg_value} cannot be cast to {_type}"
            )

        if _min is not None:
            if _min > value:
                raise argparse.ArgumentTypeError(f"Value is lower than _min = {_min}")

        if _max is not None:
            if _max < value:
                raise argparse.ArgumentTypeError(f"Value is higher than _max = {_max}")

        return value
    return validator



from starlette_web.common.management.base import BaseCommand, call_command


class Command(BaseCommand):
    help = "Command to test calling other command from code"

    async def handle(self, **options):
        await call_command("test_parser", ["1", "2", "3", "6", "--sum"])



from pathlib import Path

from starlette_web.common.conf import settings
from starlette_web.common.files.storages import MediaFileSystemStorage
from starlette_web.tests.helpers import await_


class TestFileSystemStorage:
    base_dir = settings.MEDIA["ROOT_DIR"]

    def test_fetch_file_by_url(self, client):
        rel_path = "dir1/dir2/file1.txt"

        async def write_and_get_url():
            async with MediaFileSystemStorage() as storage:
                async with storage.writer(rel_path, "t") as writer:
                    await writer.write("Test content")

                return await storage.get_url(rel_path)

        url = await_(write_and_get_url())
        assert (Path(self.base_dir) / rel_path).is_file()
        assert url == "/media/dir1/dir2/file1.txt"

        response = client.get(url)
        assert response.content.decode("utf-8") == "Test content"



import uuid

from sqlalchemy.ext.asyncio import AsyncSession

from starlette_web.contrib.auth.models import User
from starlette_web.tests.helpers import await_


def test_nested_transaction(dbs: AsyncSession):
    email = str(uuid.uuid4()).replace("-", "") + "@test.com"
    password = User.make_password(str(uuid.uuid4()))

    block1_wrapper = None
    block2_wrapper = None

    block_1_exited = False
    block_2_exited = False

    try:
        block1_wrapper = dbs.begin_nested()
        block1 = await_(block1_wrapper.__aenter__())

        block2_wrapper = dbs.begin_nested()
        block2 = await_(block2_wrapper.__aenter__())
        _ = await_(User.async_create(db_session=dbs, email=email, password=password))

        user = await_(User.async_get(db_session=dbs, email=email, password=password))
        assert user is not None
        # Note: With .begin_nested(), calling .commit()/.rollback() at end is obligatory
        await_(block2.commit())
        await_(block2_wrapper.__aexit__(None, None, None))
        block_2_exited = True

        user = await_(User.async_get(db_session=dbs, email=email, password=password))
        assert user is not None
        # Note: With .begin_nested(), calling .commit()/.rollback() at end is obligatory
        await_(block1.rollback())
        await_(block1_wrapper.__aexit__(None, None, None))
        block_1_exited = True

        user = await_(User.async_get(db_session=dbs, email=email, password=password))
        assert user is None
    finally:
        if block2_wrapper and not block_2_exited:
            await_(block2_wrapper.__aexit__(None, None, None))

        if block1_wrapper and not block_1_exited:
            await_(block1_wrapper.__aexit__(None, None, None))



from starlette_web.common.conf.base_app_config import BaseAppConfig


class AppConfig(BaseAppConfig):
    app_name = "__APPNAME"

    def initialize(self):
        pass

    def perform_checks(self):
        pass



# Copy of https://github.com/django/django/blob/main/django/db/models/enums.py
import enum

__all__ = ["Choices", "IntegerChoices", "TextChoices"]


class ChoicesMeta(enum.EnumMeta):
    """A metaclass for creating a enum choices."""

    def __new__(metacls, classname, bases, classdict):
        labels = []
        for key in classdict._member_names:
            value = classdict[key]
            label = key.replace("_", " ").title()
            labels.append(label)
            # Use dict.__setitem__() to suppress defenses against double
            # assignment in enum's classdict.
            dict.__setitem__(classdict, key, value)
        cls = super().__new__(metacls, classname, bases, classdict)
        cls._value2label_map_ = dict(zip(cls._value2member_map_, labels))
        # Add a label property to instances of enum which uses the enum member
        # that is passed in as "self" as the value to use when looking up the
        # label in the choices.
        cls.label = property(lambda self: cls._value2label_map_.get(self.value))
        cls.do_not_call_in_templates = True
        return enum.unique(cls)

    def __contains__(cls, member):
        if not isinstance(member, enum.Enum):
            # Allow non-enums to match against member values.
            return any(x.value == member for x in cls)
        return super().__contains__(member)

    @property
    def names(cls):
        empty = ["__empty__"] if hasattr(cls, "__empty__") else []
        return empty + [member.name for member in cls]

    @property
    def choices(cls):
        empty = [(None, cls.__empty__)] if hasattr(cls, "__empty__") else []
        return empty + [(member.value, member.label) for member in cls]

    @property
    def labels(cls):
        return [label for _, label in cls.choices]

    @property
    def values(cls):
        return [value for value, _ in cls.choices]


class Choices(enum.Enum, metaclass=ChoicesMeta):
    """Class for creating enumerated choices."""

    def __str__(self):
        """
        Use value when cast to str, so that Choices set as model instance
        attributes are rendered as expected in templates and similar contexts.
        """
        return str(self.value)


class IntegerChoices(int, Choices):
    """Class for creating enumerated integer choices."""

    pass


class TextChoices(str, Choices):
    """Class for creating enumerated string choices."""

    def _generate_next_value_(name, start, count, last_values):
        return name



from starlette_web.contrib.camel_case import camelize, underscoreize


def test_contrib_camelcase_camelize():
    obj = {
        "key_item": 1,
        "values_list": [
            {
                "object_id": 1,
                "object_value": "new_object_value",
            },
            {
                "object_id": 1,
                "object_value": "new_object_value",
            },
        ],
    }

    result = {
        "keyItem": 1,
        "valuesList": [
            {"objectId": 1, "objectValue": "new_object_value"},
            {"objectId": 1, "objectValue": "new_object_value"},
        ],
    }

    assert camelize(obj) == result

    obj = [
        None,
        {"item_key": "item_value"},
        "StringValue",
        False,
    ]

    result = [None, {"itemKey": "item_value"}, "StringValue", False]

    assert camelize(obj) == result


def test_contrib_camelcase_underscoreize():
    obj = {
        "keyItem": 1,
        "valuesList": [
            {
                "objectId": 1,
                "objectValue": "newObjectValue",
            },
            {
                "objectId": 2,
                "objectValue": "newObjectValue",
            },
        ],
    }

    result = {
        "key_item": 1,
        "values_list": [
            {
                "object_id": 1,
                "object_value": "newObjectValue",
            },
            {
                "object_id": 2,
                "object_value": "newObjectValue",
            },
        ],
    }

    assert underscoreize(obj) == result

    obj = [
        None,
        {"itemKey": "itemValue"},
        "StringValue",
        False,
    ]

    result = [None, {"item_key": "itemValue"}, "StringValue", False]

    assert underscoreize(obj) == result


def test_strings_are_left_as_is():
    assert camelize("a_b") == "a_b"
    assert underscoreize("aB") == "aB"



import anyio
import time
import pytest

from starlette.websockets import WebSocketDisconnect

from starlette_web.common.caches import caches
from starlette_web.contrib.auth.utils import encode_jwt
from starlette_web.tests.helpers import await_


locmem_cache = caches["locmem"]


class TestWebsocketEndpoint:
    def test_base_websocket_endpoint(self, client):
        await_(locmem_cache.async_clear())

        with client.websocket_connect("/ws/test_websocket_base") as websocket:
            cache_keys = await_(locmem_cache.async_keys("*"))
            assert cache_keys == []

            websocket.send_json({"request_type": "test"})
            time.sleep(1)
            cache_keys = await_(locmem_cache.async_keys("*"))
            assert len(cache_keys) == 1
            task_id = cache_keys[0]
            assert len(task_id) == 50

            time.sleep(2)
            cache_keys = await_(locmem_cache.async_keys("*"))
            assert len(cache_keys) == 1
            assert cache_keys[0] == task_id + "_result"
            result_value = await_(locmem_cache.async_get(cache_keys[0]))
            assert result_value == "test"

    def test_multiple_tasks_spawn(self, client):
        await_(locmem_cache.async_clear())

        with client.websocket_connect("/ws/test_websocket_base") as websocket:
            cache_keys = await_(locmem_cache.async_keys("*"))
            assert cache_keys == []

            websocket.send_json({"request_type": "test_1"})
            websocket.send_json({"request_type": "test_2"})
            websocket.send_json({"request_type": "test_3"})

            time.sleep(3)
            cache_keys = await_(locmem_cache.async_keys("*"))
            assert len(cache_keys) == 3
            results = await_(locmem_cache.async_get_many(cache_keys))
            assert set([value for key, value in results.items()]) == {"test_1", "test_2", "test_3"}

    def test_cancelled_task(self, client):
        await_(locmem_cache.async_clear())

        with client.websocket_connect("/ws/test_websocket_cancel") as websocket:
            cache_keys = await_(locmem_cache.async_keys("*"))
            assert cache_keys == []

            websocket.send_json({"request_type": "test_1"})
            websocket.send_json({"request_type": "test_2"})
            websocket.send_json({"request_type": "cancel"})

            time.sleep(3)
            cache_keys = await_(locmem_cache.async_keys("*"))
            assert len(cache_keys) == 2
            results = await_(locmem_cache.async_get_many(cache_keys))
            assert set([value for key, value in results.items()]) == {"test_1", "test_2"}

    def test_failed_task(self, client):
        await_(locmem_cache.async_clear())

        # Exception is raised by client_ws_fail,
        # due to raised Exception in 3rd background task
        with pytest.raises(WebSocketDisconnect):
            with client.websocket_connect("/ws/test_websocket_cancel") as websocket:
                cache_keys = await_(locmem_cache.async_keys("*"))
                assert cache_keys == []

                websocket.send_json({"request_type": "test_1"})
                websocket.send_json({"request_type": "test_2"})
                websocket.send_json({"request_type": "fail"})

                time.sleep(3)

        cache_keys = await_(locmem_cache.async_keys("*"))
        assert len(cache_keys) == 1
        assert cache_keys[0].endswith("_exception")
        result_value = await_(locmem_cache.async_get(cache_keys[0]))
        assert result_value == "fail"

    def test_authentication_failure(self, client):
        await_(locmem_cache.async_clear())

        with pytest.raises(WebSocketDisconnect):
            with client.websocket_connect("/ws/test_websocket_auth") as websocket:
                websocket.send_json({"request_type": "test_1"})
                time.sleep(3)

    def test_authentication_success(self, client, dbs, user, user_session):
        await_(locmem_cache.async_clear())

        jwt, _ = encode_jwt({"user_id": user.id, "session_id": user_session.public_id})
        headers = {"authorization": f"Bearer {jwt}".encode("latin-1")}

        with client.websocket_connect("/ws/test_websocket_auth", headers=headers) as websocket:
            websocket.send_json({"request_type": "test_1"})
            time.sleep(3)

            cache_keys = await_(locmem_cache.async_keys("*"))
            assert len(cache_keys) == 1
            results = await_(locmem_cache.async_get_many(cache_keys))
            assert set([value for key, value in results.items()]) == {"test_1"}

    def test_finite_periodic_task(self, client):
        await_(locmem_cache.async_clear())

        # Only async operations allow to be run with a timeout wrapper in Python
        async def _wait_for_response(ws, timeout):
            def cancellable_ws_receive_json(timeout):
                # ws.receive() does not provide a "timeout" parameter
                # without a timeout, websocket will try to get message indefinitely
                # it is also not possible to finish a thread with synchronous infinite loop
                # whereas we may not call anyio.to_process due to websocket being unpicklable
                return ws._send_queue.get(block=True, timeout=timeout)

            with anyio.fail_after(delay=timeout):
                return await anyio.to_thread.run_sync(cancellable_ws_receive_json, timeout + 1)

        with client.websocket_connect("/ws/test_websocket_finite_periodic") as websocket:
            websocket.send_json({"request_type": "periodic"})
            for i in range(4):
                response = websocket.receive_json()
                assert response["response"] == i

            # Websocket handler only sends message 4 times,
            # so next receive() should hang forever, unless one enforces timeout
            with pytest.raises(TimeoutError):
                await_(_wait_for_response(websocket, 2))

    def test_multiple_infinite_tasks(self, client):
        await_(locmem_cache.async_clear())

        with client.websocket_connect("/ws/test_websocket_infinite_periodic") as websocket:
            # Spawn 2 infinite periodic background tasks
            websocket.send_json({"request_type": "x"})
            websocket.send_json({"request_type": "y"})

            response_pool = []
            for i in range(10):
                response = websocket.receive_json()
                response_pool.append(response["response"])

            # Check, that we actually receive messages from both background tasks
            xs = [resp for resp in response_pool if resp.startswith("x_")]
            ys = [resp for resp in response_pool if resp.startswith("y_")]
            assert len(xs) >= 3
            assert len(ys) >= 3

            websocket.close(1001)

        # Assert that both tasks are properly closed after client disconnects
        time.sleep(1)
        cache_keys = await_(locmem_cache.async_keys("*"))
        assert len(cache_keys) == 2
        results = await_(locmem_cache.async_get_many(cache_keys))
        assert set([value for key, value in results.items()]) == {"finished"}



from typing import Dict, Any, List

from sqlalchemy.orm import sessionmaker

from starlette_web.common.database import make_session_maker
from starlette_web.contrib.constance.backends.base import BaseConstanceBackend
from starlette_web.contrib.constance.backends.database.models import Constance


class DatabaseBackend(BaseConstanceBackend):
    session_maker: sessionmaker

    def __init__(self):
        super().__init__()
        # Disable connection pooling to avoid creating persistent open connections
        self.session_maker = make_session_maker(use_pool=False)

    async def mget(self, keys: List[str]) -> Dict[str, Any]:
        async with self.session_maker() as session:
            constants = (await Constance.async_filter(db_session=session)).all()
            values = {key: self.empty for key in keys}
            values = {
                **values,
                **{
                    constant.key: self._preprocess_response(constant.value)
                    for constant in constants
                },
            }
            return values

    async def get(self, key: str) -> Any:
        async with self.session_maker() as session:
            val = await Constance.async_get(db_session=session, key=key)
            return self._preprocess_response(val.value if val else None)

    async def set(self, key: str, value: Any) -> None:
        async with self.session_maker() as session:
            await Constance.async_create_or_update(
                db_session=session,
                filter_kwargs={"key": key},
                update_data={"value": self.serializer.serialize(value)},
                db_commit=True,
            )



from typing import TypeVar, List, Optional

from sqlalchemy import and_, select, update, delete
from sqlalchemy.engine import ScalarResult
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.inspection import inspect
from sqlalchemy.sql import Select


class ModelMixin:
    # TODO: maybe rename "db_commit" to "commit" ?
    # TODO: add bulk methods
    """
    Base model for Sqlalchemy's ORM

    All mutable actions are flushed by default (autoflush=True, autocommit=True).
    After executing all queries, use session.commit() or session.rollback().
    Certain classes (i.e. starlette_web.common.http.base_endpoint.BaseHTTPEndpoint)
    already handle transaction logic for you.
    """

    class Meta:
        order_by = ()

    @classmethod
    def prepare_query(
        cls,
        limit: int = None,
        offset: int = None,
        order_by=(),
        **filter_kwargs,
    ) -> Select:
        _order_by = []

        if order_by:
            _order_by = [o for o in order_by]
        else:
            _order_by = [o for o in cls.Meta.order_by]

        _order_by = [
            getattr(cls, field.lstrip("-")).desc() if field.startswith("-") else getattr(cls, field)
            for field in _order_by
        ]

        query = select(cls).filter(cls._filter_criteria(filter_kwargs)).order_by(*_order_by)

        if limit is not None:
            query = query.limit(limit)

        if offset is not None:
            query = query.offset(offset)

        return query

    @classmethod
    async def async_filter(
        cls, db_session: AsyncSession, limit: int = None, offset: int = None, **filter_kwargs
    ) -> ScalarResult:
        query = cls.prepare_query(limit=limit, offset=offset, **filter_kwargs)
        result = await db_session.execute(query)
        return result.scalars()

    @classmethod
    async def async_get(cls, db_session: AsyncSession, **filter_kwargs) -> "DBModel":
        query = cls.prepare_query(**filter_kwargs)
        result = await db_session.execute(query)
        return result.scalars().first()

    @classmethod
    async def async_update(
        cls,
        db_session: AsyncSession,
        filter_kwargs: dict,
        update_data: dict,
        db_commit=False,
    ):
        query = (
            update(cls)
            .where(cls._filter_criteria(filter_kwargs))
            .values(**update_data)
            .execution_options(synchronize_session="fetch")
        )
        await db_session.execute(query)
        if db_commit:
            await db_session.commit()

    @classmethod
    async def async_delete(cls, db_session: AsyncSession, filter_kwargs: dict, db_commit=False):
        query = (
            delete(cls)
            .where(cls._filter_criteria(filter_kwargs))
            .execution_options(synchronize_session="fetch")
        )
        await db_session.execute(query)
        if db_commit:
            await db_session.commit()

    @classmethod
    async def async_create(cls, db_session: AsyncSession, db_commit=False, **data) -> "DBModel":
        instance = cls(**data)  # noqa
        db_session.add_all([instance])

        # Explicit flush() to populate newly created instance with primary key (if autogenerated).
        await db_session.flush()

        if db_commit:
            await db_session.commit()
        return instance

    @classmethod
    async def async_create_or_update(
        cls,
        db_session: AsyncSession,
        filter_kwargs: dict,
        update_data: dict,
        db_commit=False,
        force_update=False,
        update_to_null=False,
    ) -> "DBModel":
        """
        Allows to create entry, if not exists, or update an existing query.

        :param db_session: active instance of AsyncSession
        :param filter_kwargs: function will search entry by these conditions
        :param update_data: function will update these fields.
            "update_data" has priority over "filter_kwargs" in case of a new entry.
        :param db_commit: bool, whether to execute db_session.commit()
        :param force_update: bool, whether to do update is instance exists and has not been changed
        :param update_to_null: bool, whether to update currently non-null values to new null values
        :returns: instance of DBModel
        """
        # TODO: contrib.postgres.PostgresModelMixin with this method reloaded
        #  (using on_conflict do update)
        if instance := await cls.async_get(db_session, **filter_kwargs):
            if cls._object_needs_update(
                instance.to_dict(),
                update_data,
                force_update=force_update,
                update_to_null=update_to_null,
            ):
                await cls.async_update(
                    db_session,
                    filter_kwargs=filter_kwargs,
                    update_data=update_data,
                    db_commit=db_commit,
                )
        else:
            instance = await cls.async_create(
                db_session,
                **filter_kwargs,
                **update_data,
                db_commit=db_commit,
            )
        return instance

    @classmethod
    async def async_get_or_create(
        cls,
        db_session: AsyncSession,
        filter_kwargs: dict,
        extra_data: Optional[dict] = None,
        db_commit=False,
    ) -> "DBModel":
        """
        Allows to get existing instance or create, if not exists.
        Allows passing additional data upon creation.

        :param db_session: active instance of AsyncSession
        :param filter_kwargs: function will search entry by these conditions
        :param extra_data: extra data to create an entry, if not exists
        :param db_commit: bool, whether to execute db_session.commit()
        :returns: instance of DBModel
        """
        instance = await cls.async_get(db_session, **filter_kwargs)
        if not instance:
            extra_data = extra_data or {}
            instance = await cls.async_create(
                db_session,
                **filter_kwargs,
                **extra_data,
                db_commit=db_commit,
            )
        return instance

    async def update(self, db_session: AsyncSession, db_commit=False, **update_data):
        primary_keys_names = [key.name for key in inspect(self.__class__).primary_key]
        filter_kwargs = {key: getattr(self, key) for key in primary_keys_names}
        await self.async_update(db_session, filter_kwargs, update_data)

        if db_commit:
            await db_session.commit()

    async def delete(self, db_session: AsyncSession, db_commit=False):
        # The delete() method only marks object as deleted,
        # so flush() has to be used explicitly, as per documentation.
        await db_session.delete(self)
        await db_session.flush()

        if db_commit:
            await db_session.commit()

    def to_dict(self, excluded_fields: List[str] = None) -> dict:
        excluded_fields = excluded_fields or []
        res = {}
        for field in self.__dict__:
            if field not in excluded_fields and not field.startswith("_"):
                res[field] = getattr(self, field)

        return res

    @classmethod
    def _filter_criteria(cls, filter_kwargs):
        filters = []
        for filter_name, filter_value in filter_kwargs.items():
            field, _, criteria = filter_name.partition("__")
            if criteria in ("eq", ""):
                filters.append(getattr(cls, field) == filter_value)
            elif criteria == "gt":
                filters.append(getattr(cls, field) > filter_value)
            elif criteria == "lt":
                filters.append(getattr(cls, field) < filter_value)
            elif criteria == "is":
                filters.append(getattr(cls, field).is_(filter_value))
            elif criteria == "in":
                filters.append(getattr(cls, field).in_(filter_value))
            elif criteria == "notin":
                filters.append(getattr(cls, field).notin_(filter_value))
            elif criteria == "inarr":
                filters.append(getattr(cls, field).contains([filter_value]))
            elif criteria == "icontains":
                filters.append(getattr(cls, field).ilike(f"%{filter_value}%"))
            elif criteria == "ne":
                filters.append(getattr(cls, field) != filter_value)
            else:
                raise NotImplementedError(f"Unexpected criteria: {criteria}")

        return and_(True, *filters)

    @staticmethod
    def _object_needs_update(
        dict_original,
        dict_new,
        force_update=False,
        update_to_null=True,
    ) -> bool:
        if force_update:
            return True

        for key in dict_new:
            if dict_new[key] is None and not update_to_null:
                return False
            if dict_original.get(key) != dict_new[key]:
                return True

        return False


DBModel = TypeVar("DBModel", bound=ModelMixin)



import json
import datetime
from decimal import Decimal

from starlette_web.common.utils import StarletteJSONEncoder


def test_starlette_json_encoder():
    obj = {
        "key_decimal": Decimal("10.02"),
        "key_timedelta": datetime.timedelta(days=1, seconds=13876),
        "key_time": datetime.time(13, 3, 6, 999),
        "key_date": datetime.date(2020, 1, 1),
        "key_string": "string",
        "key_list": [{"key": "inside_object"}],
    }

    result = (
        '{"key_decimal": "10.02", "key_timedelta": "P1DT03H51M16S", "key_time": "13:03:06.000999", '
        '"key_date": "2020-01-01", "key_string": "string", "key_list": [{"key": "inside_object"}]}'
    )

    assert json.dumps(obj, cls=StarletteJSONEncoder) == result



import pytest

from starlette_web.common.http.exceptions import InvalidParameterError
from starlette_web.contrib.auth.models import User
from starlette_web.contrib.auth.password_validation import _get_validators, validate_password


def test_password_validation():
    assert len(_get_validators()) == 3

    with pytest.raises(InvalidParameterError) as exc:
        validate_password("123456789")
    assert exc.value.details == "Password is entirely numeric."

    with pytest.raises(InvalidParameterError) as exc:
        validate_password("am6Yab8")
    assert exc.value.details == "Password length is less than 8."

    with pytest.raises(InvalidParameterError) as exc:
        validate_password("shocking.blue", User(email="shocking.blue@gmail.com"))
    assert exc.value.details == "Password is too similar with username."

    validate_password("8IavAD9B5b81ifWU3K7G")
    validate_password("TfOqTEiDC2EezY8czZEE")
    validate_password("3pHYO6uPFkffkNN9BlPG")



import tempfile
from typing import Tuple

import anyio
import chardet


class AlembicMixin:
    @staticmethod
    async def run_alembic_main(args) -> Tuple[str, str]:
        with tempfile.NamedTemporaryFile() as stdout:
            with tempfile.NamedTemporaryFile() as stderr:
                async with await anyio.open_process(
                    ["alembic"] + args,
                    stdout=stdout,
                    stderr=stderr,
                ) as process:
                    await process.wait()

                stdout.seek(0)
                stderr.seek(0)

                out, err = stdout.read(), stderr.read()
                out_encoding = chardet.detect(out)["encoding"]
                err_encoding = chardet.detect(err)["encoding"]

                if out_encoding is None:
                    out = ""
                else:
                    out = out.decode(out_encoding)

                if err_encoding is None:
                    err = ""
                else:
                    err = err.decode(err_encoding)

                return out, err



# flake8: noqa

from starlette_web.tests.views.http import HealthCheckAPIView
from starlette_web.tests.views.websocket import (
    BaseWebsocketTestEndpoint,
    CancellationWebsocketTestEndpoint,
    AuthenticationWebsocketTestEndpoint,
    FinitePeriodicTaskWebsocketTestEndpoint,
    InfinitePeriodicTaskWebsocketTestEndpoint,
    ChatWebsocketTestEndpoint,
)



from typing import Any, List, Dict, Optional

from marshmallow.exceptions import ValidationError
from marshmallow.validate import Email as EmailValidator
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.asyncio import AsyncSession
from starlette.requests import Request
from starlette_admin.actions import action
from starlette_admin.exceptions import FormValidationError, ActionFailed

from starlette_web.contrib.admin import AdminView, admin
from starlette_web.contrib.auth.models import User, UserSession, UserInvite


class UserView(AdminView):
    exclude_fields_from_edit = ["password"]
    searchable_fields = ["email"]

    @action(
        name="delete_cascade",
        text="Delete cascade",
        confirmation="Are you sure you want to delete this items?",
        submit_btn_text="Yes, delete them all",
        submit_btn_class="btn-danger",
    )
    async def delete_cascade_action(self, request: Request, pks: List[Any]) -> str:
        affected_rows = await self.delete_users_cascade(request, pks)
        return "{} items were successfully deleted".format(affected_rows)

    @action(
        name="delete",
        text="Delete",
        confirmation="Are you sure you want to delete this items?",
        submit_btn_text="Yes, delete them all",
        submit_btn_class="btn-danger",
    )
    async def delete_action(self, request: Request, pks: List[Any]) -> str:
        try:
            affected_rows = await self.delete(request, pks)
        except SQLAlchemyError as exc:
            raise ActionFailed(str(exc))
        return "{} items were successfully deleted".format(affected_rows)

    async def validate(self, request: Request, data: Dict[str, Any]) -> None:
        # TODO: allow setting password, by hashing it here or in _populate_obj
        errors: Dict[str, str] = {}
        cleaned_email = data["email"].strip()

        try:
            EmailValidator()(cleaned_email)
        except ValidationError:
            errors["email"] = "Invalid value for field email."

        if errors:
            raise FormValidationError(errors)
        return await super().validate(request, data)

    async def delete_users_cascade(self, request: Request, pks: List[Any]) -> Optional[int]:
        session: AsyncSession = request.state.session
        objs = await self.find_by_pks(request, pks)

        async with session.begin_nested():
            for obj in objs:
                await UserSession.async_delete(
                    db_session=session,
                    filter_kwargs={"user_id": obj.id},
                )
                await UserInvite.async_delete(
                    db_session=session,
                    filter_kwargs={"user_id": obj.id},
                )
                await session.delete(obj)

        await session.commit()
        return len(objs)


class UserSessionView(AdminView):
    def can_create(self, request: Request) -> bool:
        return False

    def can_delete(self, request: Request) -> bool:
        return False


class UserInviteView(AdminView):
    def can_create(self, request: Request) -> bool:
        return False

    def can_delete(self, request: Request) -> bool:
        return False


admin.add_view(UserView(User, icon="fa fa-users"))
admin.add_view(UserSessionView(UserSession, icon="fa fa-database"))
admin.add_view(UserInviteView(UserInvite, icon="fa fa-user-plus"))



import json

from starlette_web.common.conf import settings as project_settings
from starlette_web.common.conf.base_app_config import BaseAppConfig
from starlette_web.common.http.exceptions import ImproperlyConfigured
from starlette_web.contrib.scheduler.app_settings import Settings
from starlette_web.contrib.scheduler.backends.base import JobType


class AppConfig(BaseAppConfig):
    app_name = "scheduler"

    def initialize(self):
        try:
            __import__("croniter")
        except (ImportError, SystemError):
            raise ImproperlyConfigured(
                details=(
                    "Extra dependency 'croniter' is required for "
                    "starlette_web.contrib.scheduler "
                    "Install it via 'pip install starlette-web[scheduler]'."
                )
            )

    def perform_checks(self):
        import croniter
        from croniter.croniter import CroniterError

        jobs_list = Settings(project_settings).PERIODIC_JOBS

        for job in jobs_list:
            try:
                json.JSONEncoder(sort_keys=True).encode(job)
            except (TypeError, ValueError) as exc:
                raise ImproperlyConfigured(
                    message=f"Cannot json-encode job {job}",
                    details=str(exc),
                ) from exc

            try:
                assert len(job) == 5
            except AssertionError:
                raise ImproperlyConfigured(
                    details=f"Required periodic job format in settings: {JobType}",
                )

            if job[0] != "@reboot":
                # If not @reboot, crontab pattern must be of specific form "[0] [1] [2] [3] [4]"
                # Here cron generator is created, and we yield 1 element
                # to initialize the pattern check inside croniter-lib
                try:
                    next(croniter.croniter(job[0]))
                except CroniterError as exc:
                    raise ImproperlyConfigured(details=str(exc)) from exc



import sys
from subprocess import DEVNULL

import anyio
import pytest

from starlette_web.common.conf import settings
from starlette_web.common.caches import caches
from starlette_web.common.channels.base import Channel, Event
from starlette_web.common.channels.layers.local_memory import InMemoryChannelLayer
from starlette_web.contrib.redis.channel_layers import RedisPubSubChannelLayer
from starlette_web.contrib.postgres.channel_layers import PostgreSQLChannelLayer
from starlette_web.tests.helpers import await_


class TestChannelLayers:
    # Check both correctness and idempotence of channel layers
    def test_in_memory_channel_layer(self):
        channel_ctx = Channel(InMemoryChannelLayer())
        self.run_channels_test(channel_ctx)
        self.run_channels_test(channel_ctx)
        self.run_channels_test(channel_ctx)

    def test_redis_pubsub_channel_layer(self):
        redis_options = settings.CHANNEL_LAYERS["redispubsub"]["OPTIONS"]
        channel_ctx = Channel(RedisPubSubChannelLayer(**redis_options))
        self.run_channels_test(channel_ctx)
        self.run_channels_test(channel_ctx)
        self.run_channels_test(channel_ctx)

    def test_postgres_channel_layer(self):
        psql_options = {"dsn": settings.DATABASE_DSN.replace("+asyncpg", "")}
        channel_ctx = Channel(PostgreSQLChannelLayer(**psql_options))
        self.run_channels_test(channel_ctx)
        self.run_channels_test(channel_ctx)
        self.run_channels_test(channel_ctx)

    def run_channels_test(self, channel_ctx: Channel):
        accepted_messages = []

        async def publisher_task(channel):
            with anyio.fail_after(2):
                for i in range(5):
                    await anyio.sleep(0.2)
                    await channel.publish("test_group", f"Message {i}")

            # Tests utilize subscribers in an infinite-loop manner,
            # so an outer exception must be raised in order for it to stop.
            # In real application subscribers are expected to be in Websockets,
            # which receive WebsocketDisconnected exception when finished,
            # causing subscriber to stop.
            # This is a mock for tests.
            # Exception will propagate to task group and
            # will close infinite consumers as well.
            await anyio.sleep(0.1)
            raise Exception

        async def subscriber_task(channel: Channel, messages_pool):
            async with channel.subscribe("test_group") as subscriber:
                async for message in subscriber:
                    messages_pool.append(message)

        async def task_coroutine():
            nonlocal accepted_messages

            async with channel_ctx as channels:
                async with anyio.create_task_group() as task_group:
                    task_group.start_soon(publisher_task, channels)
                    task_group.start_soon(subscriber_task, channels, accepted_messages)
                    task_group.start_soon(subscriber_task, channels, accepted_messages)
                    task_group.start_soon(subscriber_task, channels, accepted_messages)

        # Redis will actually raise ExceptionGroup,
        # having 3 subscribers within task group
        # that will raise ConnectionError after close
        with pytest.raises(Exception):
            await_(task_coroutine())

        assert len(accepted_messages) == 15
        messages = []
        for event in accepted_messages:
            assert type(event) == Event
            messages.append(event.message)

        for i in range(5):
            assert messages.count(f"Message {i}") == 3

    def test_channels_crossprocess(self):
        test_group_name = "6B65mbrpcAy4zHAjNOZW8T84c7Yl2b"
        subscriber_1 = "Ej3Qu6JFtVbUXh9qb20z"
        subscriber_2 = "Eh9DnQeYvlLI2x5lkMPb"
        subscriber_3 = "ooF5oCaGYAE9dy8aXBSI"

        async def run_command_in_process(command_name, command_args):
            cmd = (
                f"cd {settings.PROJECT_ROOT_DIR} && "
                f"{sys.executable} command.py {command_name}"
            )

            if command_args:
                cmd += " " + " ".join(command_args)

            async with await anyio.open_process(
                cmd, stdin=DEVNULL, stdout=DEVNULL, stderr=DEVNULL
            ) as process:
                await process.wait()

        async def task_coroutine():
            async with anyio.create_task_group() as nursery:
                nursery.cancel_scope.deadline = anyio.current_time() + 10
                nursery.start_soon(
                    run_command_in_process,
                    "test_channels_publisher",
                    [f"--group={test_group_name}"],
                )
                nursery.start_soon(
                    run_command_in_process,
                    "test_channels_subscriber",
                    [f"--group={test_group_name}", f"--subscriber={subscriber_1}"],
                )
                nursery.start_soon(
                    run_command_in_process,
                    "test_channels_subscriber",
                    [f"--group={test_group_name}", f"--subscriber={subscriber_2}"],
                )
                nursery.start_soon(
                    run_command_in_process,
                    "test_channels_subscriber",
                    [f"--group={test_group_name}", f"--subscriber={subscriber_3}"],
                )

        await_(task_coroutine())

        # TODO: share data via file-cache in tests
        default_cache = caches["default"]
        publisher_flag = await_(default_cache.async_get(f"{test_group_name}-publisher-done"))
        assert publisher_flag == 1

        subscriber_1_flag = await_(
            default_cache.async_get(f"{test_group_name}-{subscriber_1}-done")
        )
        subscriber_2_flag = await_(
            default_cache.async_get(f"{test_group_name}-{subscriber_2}-done")
        )
        subscriber_3_flag = await_(
            default_cache.async_get(f"{test_group_name}-{subscriber_3}-done")
        )
        assert subscriber_1_flag == ["Message 0", "Message 1", "Message 2", "DONE"]
        assert subscriber_2_flag == ["Message 0", "Message 1", "Message 2", "DONE"]
        assert subscriber_3_flag == ["Message 0", "Message 1", "Message 2", "DONE"]

    def test_immediate_subscriber_cleanup_on_exit(self):
        async def task_coroutine():
            _publisher_result = []

            async def publisher_task(channel, _res):
                await channel.publish("test_group", "Message")
                await anyio.sleep(0.2)
                _res.append(len(channel._subscribers))
                _res.append(len(channel._channel_layer._subscribed))

            async def subscriber_task(channel: Channel):
                async with channel.subscribe("test_group") as subscriber:
                    async for message in subscriber:
                        # break immediately
                        break

            async with Channel(InMemoryChannelLayer()) as channels:
                async with anyio.create_task_group() as task_group:
                    task_group.start_soon(publisher_task, channels, _publisher_result)
                    task_group.start_soon(subscriber_task, channels)
                    task_group.start_soon(subscriber_task, channels)
                    task_group.start_soon(subscriber_task, channels)

            return _publisher_result

        res = await_(task_coroutine())
        assert res == [0, 0]



import anyio

from redis import asyncio as aioredis
from redis.asyncio.lock import Lock as AioredisLock

from starlette_web.common.caches.base import CacheLockError


class RedisLock(AioredisLock):
    EXIT_MAX_DELAY: float = 60

    async def __aenter__(self):
        try:
            return await super().__aenter__()
        except aioredis.RedisError as exc:
            raise CacheLockError from exc

    async def __aexit__(self, *args):
        try:
            # Do not raise exception, if lock has been released
            # (might have been released due to timeout)
            expected_token = self.local.token
            if expected_token is not None:
                self.local.token = None

                async def close_task():
                    await self.lua_release(
                        keys=[self.name],
                        args=[expected_token],
                        client=self.redis,
                    )

                with anyio.fail_after(self.EXIT_MAX_DELAY, shield=True):
                    await close_task()
        except (aioredis.RedisError, TimeoutError) as exc:
            raise CacheLockError from exc



from contextlib import asynccontextmanager
from typing import AsyncGenerator, AsyncIterator, Optional, Any, Dict, Set

import anyio
from anyio._core._tasks import TaskGroup
from anyio.streams.memory import (
    MemoryObjectReceiveStream,
    MemoryObjectSendStream,
    EndOfStream,
    ClosedResourceError,
)

from starlette_web.common.channels.layers.base import BaseChannelLayer
from starlette_web.common.channels.event import Event
from starlette_web.common.channels.exceptions import ListenerClosed


_empty = object()


class Channel:
    EXIT_MAX_DELAY = 60

    def __init__(self, channel_layer: BaseChannelLayer):
        self._task_group: Optional[TaskGroup] = None
        self._channel_layer = channel_layer
        self._subscribers: Dict[str, Set[MemoryObjectSendStream]] = dict()
        self._manager_lock = anyio.Lock()
        self._task_group_handler = None

    async def __aenter__(self) -> "Channel":
        await self.connect()
        self._task_group_handler = anyio.create_task_group()
        self._task_group = await self._task_group_handler.__aenter__()
        self._task_group.start_soon(self._listener)
        return self

    async def __aexit__(self, *args: Any, **kwargs: Any):
        try:
            self._task_group.cancel_scope.cancel()
            del self._task_group

            retval = await self._task_group_handler.__aexit__(*args)
            del self._task_group_handler
        finally:
            self._subscribers.clear()
            with anyio.fail_after(self.EXIT_MAX_DELAY, shield=True):
                await self.disconnect()

        return retval

    async def connect(self) -> None:
        await self._channel_layer.connect()

    async def disconnect(self) -> None:
        await self._channel_layer.disconnect()

    async def _listener(self) -> None:
        while True:
            try:
                event = await self._channel_layer.next_published()
            except ListenerClosed:
                break

            async with self._manager_lock:
                subscribers_list = list(self._subscribers.get(event.group, []))

            async with anyio.create_task_group() as nursery:
                for send_stream in subscribers_list:
                    nursery.start_soon(send_stream.send, event)

        async with self._manager_lock:
            for group in self._subscribers.keys():
                for recv_channel in self._subscribers[group]:
                    recv_channel.close()

    async def publish(self, group: str, message: Any) -> None:
        await self._channel_layer.publish(group, message)

    @asynccontextmanager
    async def subscribe(self, group: str) -> AsyncGenerator["Subscriber", None]:
        send_stream, receive_stream = anyio.create_memory_object_stream()

        try:
            async with self._manager_lock:
                if not self._subscribers.get(group):
                    await self._channel_layer.subscribe(group)
                    self._subscribers[group] = {
                        send_stream,
                    }
                else:
                    self._subscribers[group].add(send_stream)

            yield Subscriber(receive_stream)

        finally:
            try:
                with anyio.fail_after(self.EXIT_MAX_DELAY, shield=True):
                    async with self._manager_lock:
                        self._subscribers[group].remove(send_stream)
                        if not self._subscribers.get(group):
                            del self._subscribers[group]
                            await self._channel_layer.unsubscribe(group)

            finally:
                send_stream.close()


class Subscriber:
    def __init__(self, receive_stream: MemoryObjectReceiveStream) -> None:
        self._receive_stream = receive_stream

    async def __aiter__(self) -> AsyncIterator[Event]:
        async with self._receive_stream:
            try:
                while True:
                    event: Event = await self._receive_stream.receive()
                    yield event
            except (EndOfStream, ClosedResourceError):
                pass



import hashlib
import uuid

from starlette_web.contrib.auth.hashers import (
    PBKDF2PasswordHasher,
    make_password,
    verify_password,
    _password_manager,
    BasePasswordHasher,
)


class MD5TestPasswordHasher(BasePasswordHasher):
    """
    The Salted MD5 password hashing algorithm (not recommended)
    """

    algorithm = "md5"

    def encode(self, password, salt=None) -> str:
        assert password is not None
        assert salt and "$" not in salt
        hash = hashlib.md5((salt + password).encode()).hexdigest()
        return "%s$1$%s$%s" % (self.algorithm, salt, hash)


def test_password_encode():
    hasher = PBKDF2PasswordHasher()
    raw_password = uuid.uuid4().hex
    encoded = hasher.encode(raw_password, salt="test_salt")
    algorithm, iterations, salt, hash_ = encoded.split("$", 3)
    assert salt == "test_salt"
    assert algorithm == "pbkdf2_sha256"
    assert hash_ != raw_password


def test_password_verify__ok():
    hasher = PBKDF2PasswordHasher()
    raw_password = uuid.uuid4().hex
    encoded = hasher.encode(raw_password, salt="test_salt")
    verified = hasher.verify(raw_password, encoded)
    assert verified


def test_password_verify__incompatible_format__fail():
    hasher = PBKDF2PasswordHasher()
    verified = hasher.verify(uuid.uuid4().hex, "fake-encoded-password")
    assert not verified


def test_password_verify__algorithm_mismatch__fail():
    hasher = PBKDF2PasswordHasher()
    verified = hasher.verify(uuid.uuid4().hex, "fake-algorithm$1000$salt$enc-password")
    assert not verified


def test_generic_hasher_utils():
    hasher = PBKDF2PasswordHasher()
    raw_password = uuid.uuid4().hex
    salt = hasher.salt()

    password_1 = hasher.encode(raw_password, salt)
    password_2 = make_password(raw_password, salt)
    assert password_1 == password_2
    assert verify_password(raw_password, password_1)


def test_additional_hasher():
    _password_manager._add_password_hasher(MD5TestPasswordHasher())
    raw_password = uuid.uuid4().hex
    salt = uuid.uuid4().hex

    # Python 3.7+ guarantees order of keys, so PBKDF2PasswordHasher is always first
    # as long as it is first in settings.AUTH_PASSWORD_HASHERS
    # https://mail.python.org/pipermail/python-dev/2017-December/151283.html
    encoded_password = make_password(raw_password, salt)
    pdkdf2_password = PBKDF2PasswordHasher().encode(raw_password, salt)
    assert encoded_password == pdkdf2_password

    assert PBKDF2PasswordHasher().verify(raw_password, encoded_password)
    assert verify_password(raw_password, encoded_password)

    raw_password = uuid.uuid4().hex
    salt = uuid.uuid4().hex
    md5_password = MD5TestPasswordHasher().encode(raw_password, salt)
    encoded_password = make_password(raw_password, salt)
    assert md5_password != encoded_password
    assert verify_password(raw_password, md5_password)



# Adapted from https://github.com/django/django/blob/main/django/core/cache/backends/locmem.py

import math
import re
from typing import Any, Optional, Dict, Sequence, AsyncContextManager, List

import anyio
from anyio.lowlevel import checkpoint

from starlette_web.common.caches.base import BaseCache, CacheError
from starlette_web.common.caches.base_lock import BaseLock
from starlette_web.common.http.exceptions import ImproperlyConfigured
from starlette_web.common.utils.regex import redis_pattern_to_re_pattern


_caches: Dict[str, Dict[str, Any]] = {}
_expire_info: Dict[str, Dict[str, float]] = {}
_locks: Dict[str, Dict[str, float]] = {}


class _AsyncLocalMemoryLock(BaseLock):
    def __init__(
        self,
        name: str,
        timeout: Optional[float] = None,
        blocking_timeout: Optional[float] = None,
        **kwargs,
    ) -> None:
        super().__init__(name=name, timeout=timeout, blocking_timeout=blocking_timeout, **kwargs)
        self._manager_lock = kwargs["manager_lock"]
        self._cache_name = kwargs["cache_name"]
        global _locks
        _locks.setdefault(self._cache_name, {})
        self._cache_lock = _locks[self._cache_name]

    async def _acquire(self):
        if self._is_acquired:
            return

        while True:
            await checkpoint()
            async with self._manager_lock:
                if self._cache_lock.get(self._name, -1) < anyio.current_time():
                    self._cache_lock[self._name] = anyio.current_time() + self._timeout
                    self._acquire_event.set()
                    return

    async def _release(self):
        if not self._is_acquired:
            return

        async with self._manager_lock:
            self._cache_lock[self._name] = -1
            self._is_acquired = False


class LocalMemoryCache(BaseCache):
    def __init__(self, options):
        self.name = options.get("name", None)
        if self.name is None:
            raise ImproperlyConfigured('LocalMemoryCache must be instantiated with option "name"')

        super().__init__(options)
        self._manager_lock = anyio.Lock()
        self._locking_manager_lock = anyio.Lock()

        global _caches, _locks, _expire_info
        self._cache = _caches.setdefault(self.name, {})
        self._expire_info = _expire_info.setdefault(self.name, {})
        self._lock = _locks.setdefault(self.name, {})

    async def async_get(self, key: str) -> Any:
        async with self._manager_lock:
            if self._has_expired(key):
                self._delete_key(key)

            return self.serializer.deserialize(self._cache.get(key))

    async def async_set(self, key: str, value: Any, timeout: Optional[float] = 120) -> None:
        async with self._manager_lock:
            deadline = anyio.current_time() + timeout if timeout is not None else math.inf
            self._cache[key] = self.serializer.serialize(value)
            self._expire_info[key] = deadline

    async def async_delete(self, key: str) -> None:
        async with self._manager_lock:
            self._delete_key(key)

    async def async_keys(self, pattern: str) -> List[str]:
        try:
            re_pattern = redis_pattern_to_re_pattern(pattern)
        except re.error as exc:
            raise CacheError(details=str(exc)) from exc

        return [
            key
            for key in self._cache.keys()
            if not self._has_expired(key) and re.fullmatch(re_pattern, key)
        ]

    async def async_has_key(self, key: str) -> bool:
        async with self._manager_lock:
            if self._has_expired(key):
                self._delete_key(key)

            return key in self._cache

    def _delete_key(self, key: str) -> None:
        self._cache.pop(key, None)
        self._expire_info.pop(key, None)

    def _has_expired(self, key) -> bool:
        return self._expire_info.get(key, -1) < anyio.current_time()

    async def async_get_many(self, keys: Sequence[str]) -> Dict[str, Any]:
        result = dict()
        for key in keys:
            result[key] = await self.async_get(key)
        return result

    async def async_set_many(self, data: Dict[str, Any], timeout: Optional[float] = 120) -> None:
        for key, value in data.items():
            await self.async_set(key, value, timeout=timeout)

    async def async_delete_many(self, keys: Sequence[str]) -> None:
        async with self._manager_lock:
            for key in keys:
                self._delete_key(key)

    async def async_clear(self) -> None:
        async with self._manager_lock:
            self._expire_info.clear()
            self._cache.clear()

    def lock(
        self,
        name: str,
        timeout: Optional[float] = 20.0,
        blocking_timeout: Optional[float] = None,
        **kwargs,
    ) -> AsyncContextManager:
        return _AsyncLocalMemoryLock(
            name=name,
            timeout=timeout,
            blocking_timeout=blocking_timeout,
            manager_lock=self._locking_manager_lock,
            cache_name=self.name,
            **kwargs,
        )



from starlette_web.common.caches import caches
from starlette_web.tests.core.helpers.base_cache_tester import BaseCacheTester


class TestFileCache(BaseCacheTester):
    def test_file_cache_base_ops(self):
        self._run_base_cache_test(caches["files"])

    def test_file_cache_many_ops(self):
        self._run_cache_many_ops_test(caches["files"])

    def test_file_lock(self):
        self._run_cache_lock_test(caches["files"])

    def test_file_lock_race_condition(self):
        self._run_cache_mutual_lock_test(caches["files"])

    def test_file_lock_correct_task_blocking(self):
        self._run_locks_timeouts_test(caches["files"])

    def test_file_lock_cancellation(self):
        self._run_base_lock_cancellation(caches["files"])


class TestInMemoryCache(BaseCacheTester):
    def test_locmem_cache_base_ops(self):
        self._run_base_cache_test(caches["locmem"])

    def test_locmem_cache_many_ops(self):
        self._run_cache_many_ops_test(caches["locmem"])

    def test_locmem_lock(self):
        self._run_cache_lock_test(caches["locmem"])

    def test_locmem_lock_race_condition(self):
        self._run_cache_mutual_lock_test(caches["locmem"])

    def test_locmem_lock_correct_task_blocking(self):
        self._run_locks_timeouts_test(caches["locmem"])

    def test_file_lock_cancellation(self):
        self._run_base_lock_cancellation(caches["locmem"])



import base64
import hashlib
import math
from typing import Optional, Dict, Union

from starlette_web.common.conf import settings
from starlette_web.common.http.exceptions import ImproperlyConfigured
from starlette_web.common.utils.importing import import_string
from starlette_web.common.utils.singleton import Singleton
from starlette_web.common.utils.crypto import (
    get_random_string,
    constant_time_compare,
    RANDOM_STRING_CHARS,
)


class BasePasswordHasher(metaclass=Singleton):
    algorithm = None
    salt_entropy = 128

    def salt(self) -> str:
        char_count = math.ceil(self.salt_entropy / math.log2(len(RANDOM_STRING_CHARS)))
        return get_random_string(char_count, allowed_chars=RANDOM_STRING_CHARS)

    def encode(self, password: str, salt: Optional[str] = None) -> str:
        raise NotImplementedError

    def verify(self, password: str, encoded: str) -> bool:
        """Check if the given password is correct."""
        try:
            algorithm, iterations, salt, _ = encoded.split("$", 3)
        except ValueError:
            return False

        if algorithm != self.algorithm:
            return False

        encoded_2 = self.encode(password, salt)
        return constant_time_compare(encoded, encoded_2)


class PBKDF2PasswordHasher(BasePasswordHasher):
    """
    Secure password hashing using the PBKDF2 algorithm (recommended)

    Configured to use PBKDF2 + HMAC + SHA256.
    The result is a 64 byte binary string. Iterations may be changed
    safely, but you must rename the algorithm if you change SHA256.
    """

    algorithm = "pbkdf2_sha256"
    iterations = 180000
    digest = hashlib.sha256

    def encode(self, password: str, salt: str = None) -> str:
        """Encoding password using random salt + pbkdf2_sha256"""
        salt = salt or self.salt()
        assert password is not None
        assert salt and "$" not in salt
        hash_ = self._pbkdf2(password, salt)
        hash_ = base64.b64encode(hash_).decode("ascii").strip()
        return "%s$%d$%s$%s" % (self.algorithm, self.iterations, salt, hash_)

    def _pbkdf2(self, password: str, salt: str) -> bytes:
        """Return the hash of password using pbkdf2."""
        digest = self.digest
        iterations = self.iterations
        password = bytes(password, encoding="utf-8")
        salt = bytes(salt, encoding="utf-8")
        return hashlib.pbkdf2_hmac(digest().name, password, salt, iterations)


class PasswordManager(metaclass=Singleton):
    def __init__(self):
        self._password_hashers: Dict[str, BasePasswordHasher] = dict()
        for password_hasher in settings.AUTH_PASSWORD_HASHERS:
            self._add_password_hasher(password_hasher)

    def _add_password_hasher(self, hasher: Union[str, BasePasswordHasher]) -> None:
        if type(hasher) == str:
            self._password_hashers[hasher] = import_string(hasher)()
        else:
            self._password_hashers[hasher.__class__.__name__] = hasher

    def verify_password(self, password: str, encoded: str) -> bool:
        return any(
            (
                self._password_hashers[hasher].verify(password, encoded)
                for hasher in self._password_hashers.keys()
            )
        )

    def make_password(self, password: str, salt: Optional[str] = None) -> str:
        try:
            hasher = settings.AUTH_PASSWORD_HASHERS[0]
        except IndexError:
            raise ImproperlyConfigured(
                "At least 1 password hasher must be defined at settings.AUTH_PASSWORD_HASHERS"
            )

        return self._password_hashers[hasher].encode(password, salt)

    def get_hashers(self) -> Dict[str, BasePasswordHasher]:
        return self._password_hashers


_password_manager = PasswordManager()
make_password = _password_manager.make_password
verify_password = _password_manager.verify_password



# flake8: noqa

from pathlib import Path

from starlette.config import Config
from starlette.datastructures import Secret

PROJECT_ROOT_DIR = Path(__file__).parent.parent

config = Config(PROJECT_ROOT_DIR / ".env")

APP_DEBUG = config("APP_DEBUG", cast=bool, default=False)
SECRET_KEY = config("SECRET_KEY", cast=Secret, default="project-secret")
SITE_URL = "https://web.project.com"

INSTALLED_APPS = [
    "starlette_web.contrib.staticfiles",

    # Following contrib modules require extra dependencies.
    # "starlette_web.contrib.apispec",
    # "starlette_web.contrib.auth",
    # "starlette_web.contrib.admin",
    # "starlette_web.contrib.constance",
    # "starlette_web.contrib.constance.backends.database",
    # "starlette_web.contrib.scheduler",
]

DB_ECHO = config("DB_ECHO", cast=bool, default=False)
DB_NAME = config("DB_NAME", default="web_project")

DATABASE = {
    "driver": "postgresql+asyncpg",
    "host": config("DB_HOST", default="localhost"),
    "port": config("DB_PORT", cast=int, default=5432),
    "username": config("DB_USERNAME", default="starlette-web"),
    "password": config("DB_PASSWORD", cast=Secret, default=None),
    "database": DB_NAME,
    "pool_min_size": config("DB_POOL_MIN_SIZE", cast=int, default=1),
    "pool_max_size": config("DB_POOL_MAX_SIZE", cast=int, default=16),
    "ssl": config("DB_SSL", default=None),
    "use_connection_for_request": config("DB_USE_CONNECTION_FOR_REQUEST", cast=bool, default=True),
    "retry_limit": config("DB_RETRY_LIMIT", cast=int, default=1),
    "retry_interval": config("DB_RETRY_INTERVAL", cast=int, default=1),
}

DATABASE_DSN = config(
    "DB_DSN",
    cast=str,
    default="{driver}://{username}:{password}@{host}:{port}/{database}".format(**DATABASE),
)

ROUTES = "core.routes.routes"

CACHES = {
    "default": {
        "BACKEND": "starlette_web.common.caches.local_memory.LocalMemoryCache",
        "OPTIONS": {
            "name": "default",
        },
    },
}

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {
            "format": "[%(asctime)s] %(levelname)s [%(name)s:%(lineno)s] %(message)s",
            "datefmt": "%d.%m.%Y %H:%M:%S",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "standard",
        },
    },
    "loggers": {
        "uvicorn": {
            "handlers": ["console"],
            "level": "INFO",
            "propagate": False,
        },
        "starlette_web": {
            "handlers": ["console"],
            "level": "INFO",
            "propagate": False,
        },
    },
}

TEMPLATES = {
    "ROOT_DIR": PROJECT_ROOT_DIR / "templates",
    "AUTOESCAPE": False,
    "AUTORELOAD": False,
}

STATIC = {
    "ROOT_DIR": PROJECT_ROOT_DIR / "static",
    "URL": "/static",
}
STATIC["ROOT_DIR"].mkdir(exist_ok=True)

MEDIA = {
    "ROOT_DIR": PROJECT_ROOT_DIR / "media",
    "URL": "/media/",
}
MEDIA["ROOT_DIR"].mkdir(exist_ok=True)

APISPEC = {
    "CONFIG": dict(
        title="Project documentation",
        version="0.0.1",
        openapi_version="3.0.2",
        info=dict(description="My custom project."),
    ),
    "CONVERT_TO_CAMEL_CASE": False,
}



# flake8: noqa

from starlette_web.common.utils.choices import TextChoices, IntegerChoices
from starlette_web.common.utils.crypto import get_random_string, constant_time_compare
from starlette_web.common.utils.importing import import_string
from starlette_web.common.utils.inspect import get_available_options, safe_init
from starlette_web.common.utils.json import StarletteJSONEncoder
from starlette_web.common.utils.singleton import Singleton
from starlette_web.common.utils.urls import urljoin



from sqlalchemy import types


class ChoiceType(types.TypeDecorator):
    """
    This implementation taken from https://github.com/kvesteri/sqlalchemy-utils/

    Columns with ChoiceTypes are automatically coerced to Choice objects while
    a list of tuple been passed to the constructor. If a subclass of
    :class:`enum.Enum` is passed, columns will be coerced to value of :class:`enum.Enum`

    >>> import enum
    >>> from sqlalchemy import Column, String
    >>> from starlette_web.common.database import ModelBase

    >>> class UserType(enum.Enum):
    >>>     admin = 'admin'
    >>>     regular = 'regular'

    >>> class User(ModelBase):
    >>>     __tablename__ = 'user'
    >>>     type = Column(ChoiceType(UserType, impl=String(16)))

    >>> user = User(type='admin')
    >>> user.type
    [0] 'admin'

    """

    impl = types.String(255)
    cache_ok = True

    def __init__(self, enum_class, impl=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.enum_class = enum_class
        if impl:
            self.impl = impl() if isinstance(impl, type) else impl

    def __repr__(self):
        """
        It is used for representation type in alembic logic (render result in mako template)
        migrations.env.render_item uses this representation as "sa.%r" % obj. ex.: "sa.VARCHAR"
        """
        return str(self.impl)

    @property
    def python_type(self):
        return self.impl.python_type

    def coercion_listener(self, target, value, oldvalue, initiator):  # noqa
        return self._coerce(value)

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return self.enum_class(value).value

    def process_result_value(self, value, dialect):
        return self._coerce(value)

    def process_literal_param(self, value, dialect):
        return ""

    def _coerce(self, value):
        if value is None:
            return None
        return self.enum_class(value)



from traceback_with_variables import format_exc

from starlette_web.common.conf import settings
from starlette_web.common.conf.base_app_config import BaseAppConfig
from starlette_web.common.http.exceptions import ImproperlyConfigured
from starlette_web.common.utils import import_string
from starlette_web.contrib.apispec.views import schemas


class AppConfig(BaseAppConfig):
    app_name = "apispec"

    def initialize(self):
        try:
            __import__("openapi_spec_validator")
        except (SystemError, ImportError):
            raise ImproperlyConfigured(
                details=(
                    "Extra dependency 'openapi_spec_validator' is required"
                    " for starlette_web.contrib.apispec "
                    "Install it via 'pip install starlette-web[apispec]'."
                )
            )

    def perform_checks(self):
        from openapi_spec_validator import validate_spec
        from openapi_spec_validator.validation.exceptions import (
            OpenAPIValidationError,
            OpenAPISpecValidatorError,
        )

        routes = import_string(settings.ROUTES)

        try:
            # This check mostly fails on invalid indentation
            # or partially missing properties.
            api_spec = schemas.get_schema(routes)
        except Exception as exc:  # noqa
            # Printing variable values in traceback is the only
            # viable way to know which schema is fallible
            raise ImproperlyConfigured(
                message="Invalid schema in apispec",
                details=format_exc(exc, num_skipped_frames=0),
            )

        try:
            # This check finds missing whole blocks,
            # i.e. missing info about path parameter in schema
            validate_spec(api_spec)
        except (OpenAPIValidationError, OpenAPISpecValidatorError) as exc:
            raise ImproperlyConfigured(
                details=str(exc),
            )



import re
import pytest

from starlette_web.common.utils.regex import redis_pattern_to_re_pattern


def test_redis_pattern_matcher():
    re_pattern = redis_pattern_to_re_pattern("keys:*")
    assert re.fullmatch(re_pattern, "keys:")
    assert re.fullmatch(re_pattern, "keys:1")
    assert re.fullmatch(re_pattern, "keys:abcde")
    assert re.fullmatch(re_pattern, "key:abcde") is None

    re_pattern = redis_pattern_to_re_pattern("keys:??")
    assert re.fullmatch(re_pattern, "keys:\\\\")
    assert re.fullmatch(re_pattern, "keys:3a")
    assert re.fullmatch(re_pattern, "keys:??")
    assert re.fullmatch(re_pattern, "keys:3aa") is None

    re_pattern = redis_pattern_to_re_pattern("h[^e]llo")
    assert re.fullmatch(re_pattern, "hello") is None
    assert re.fullmatch(re_pattern, "hallo")
    assert re.fullmatch(re_pattern, "hxllo")

    re_pattern = redis_pattern_to_re_pattern("h[^e-z]llo")
    assert re.fullmatch(re_pattern, "hello") is None
    assert re.fullmatch(re_pattern, "hzllo") is None
    assert re.fullmatch(re_pattern, "hallo")
    assert re.fullmatch(re_pattern, "hAllo")

    re_pattern = redis_pattern_to_re_pattern("h[a-b]llo")
    assert re.fullmatch(re_pattern, "hallo")
    assert re.fullmatch(re_pattern, "hbllo")
    assert re.fullmatch(re_pattern, "hello") is None


def test_invalid_patterns():
    with pytest.raises(re.error):
        redis_pattern_to_re_pattern("[")



# Modification of https://github.com/Woile/starlette-apispec/blob/master/starlette_apispec/schemas.py  # noqa: E501
from collections import deque
from typing import List

from apispec import APISpec
from apispec.exceptions import DuplicateComponentNameError
from starlette.routing import BaseRoute, Mount, Route
from starlette.schemas import BaseSchemaGenerator, EndpointInfo

from starlette_web.common.authorization.backends import BaseAuthenticationBackend
from starlette_web.common.http.schemas import get_error_schema_class
from starlette_web.common.i18n import gettext


# TODO: custom EndpointInfo with link to endpoint class
# TODO: maybe allow user to override class in settings (?)
class APISpecSchemaGenerator(BaseSchemaGenerator):
    ERROR_SCHEMA_NAME = "Error"

    def __init__(self, spec: APISpec) -> None:
        self.spec: APISpec = spec
        self.security_schemas = {}
        self.paths_and_endpoints = {}

    def _fetch_paths_and_endpoints(self, initial_routes: List[BaseRoute]):
        if self.paths_and_endpoints:
            return self.paths_and_endpoints

        mounts = deque([])
        routes = {}

        for route in initial_routes:
            # Skip any websocket connections
            if isinstance(route, Mount):
                mounts.append((self._remove_converter(route.path), route))
            elif isinstance(route, Route) and getattr(route, "include_in_schema", True):
                routes[self._remove_converter(route.path)] = route.endpoint

        while mounts:
            path, mount = mounts.popleft()
            for route in mount.routes:
                # Skip any websocket connections
                if isinstance(route, Mount):
                    mounts.append((path + self._remove_converter(route.path), route))
                elif isinstance(route, Route) and getattr(route, "include_in_schema", True):
                    routes[path + self._remove_converter(route.path)] = route.endpoint

        self.paths_and_endpoints = routes

    def _add_security_schema(self, auth_backend: BaseAuthenticationBackend) -> str:
        auth_name = getattr(auth_backend, "openapi_name", None) or auth_backend.__name__

        if auth_name in self.security_schemas:
            if self.security_schemas[auth_name] != auth_backend:
                raise DuplicateComponentNameError(
                    gettext("Duplicate auth_backend $auth_name", auth_name=auth_name)
                )
        else:
            self.security_schemas[auth_name] = auth_backend
            self.spec.components.security_scheme(auth_name, auth_backend.openapi_spec)

        return auth_name

    def _populate_security_schema(self, endpoint: EndpointInfo, parsed_docstring: dict):
        if parsed_docstring.get("security"):
            return

        kls = self.paths_and_endpoints[endpoint.path]
        if not getattr(kls, "auth_backend", None):
            return

        if getattr(kls.auth_backend, "openapi_spec", None):
            security_schema = self._add_security_schema(kls.auth_backend)
            # TODO: examine, if something should be passed into dict object
            parsed_docstring["security"] = [{security_schema: []}]

    def _populate_auth_errors(self, endpoint: EndpointInfo, parsed_docstring: dict):
        kls = self.paths_and_endpoints[endpoint.path]

        if getattr(kls, "auth_backend", None):
            parsed_docstring["responses"] = {
                **parsed_docstring.get("responses", {}),
                "401": {
                    "description": gettext("Authentication error."),
                    "content": {"application/json": {"schema": self.ERROR_SCHEMA_NAME}},
                },
            }

        if getattr(kls, "permission_classes", []):
            parsed_docstring["responses"] = {
                **parsed_docstring.get("responses", {}),
                "403": {
                    "description": gettext("Access forbidden."),
                    "content": {"application/json": {"schema": self.ERROR_SCHEMA_NAME}},
                },
            }

    def _populate_validation(self, endpoint: EndpointInfo, parsed_docstring: dict):
        if "requestBody" in parsed_docstring:
            parsed_docstring["responses"] = {
                **parsed_docstring.get("responses", {}),
                "400": {
                    "description": gettext("Bad Request."),
                    "content": {"application/json": {"schema": self.ERROR_SCHEMA_NAME}},
                },
            }

    def get_schema(self, routes: List[BaseRoute]) -> dict:
        ErrorResponseSchema = get_error_schema_class()()
        if self.ERROR_SCHEMA_NAME not in self.spec.components.schemas:
            self.spec.components.schema(self.ERROR_SCHEMA_NAME, schema=ErrorResponseSchema)

        endpoints = self.get_endpoints(routes)
        self._fetch_paths_and_endpoints(routes)

        for endpoint in endpoints:
            parsed = self.parse_docstring(endpoint.func)
            self._populate_security_schema(endpoint, parsed)
            self._populate_auth_errors(endpoint, parsed)
            self._populate_validation(endpoint, parsed)

            self.spec.path(
                path=endpoint.path,
                operations={endpoint.http_method: parsed},
            )

        return self.spec.to_dict()



from importlib import import_module
from pathlib import Path
import shutil

from starlette_web.common.conf import settings
from starlette_web.common.management.base import BaseCommand, CommandError


# TODO: enhance variability
# TODO: handle errors
# TODO: probably need refactoring ?
class Command(BaseCommand):
    help = (
        "Move static files and templates from INSTALLED_APPS "
        "directories to static root and templates root, respectively"
    )
    SKIP_APPS = ["starlette_web.common", "starlette_web.contrib.admin"]

    async def handle(self, **options):
        Path.mkdir(settings.TEMPLATES["ROOT_DIR"], exist_ok=True)
        Path.mkdir(settings.STATIC["ROOT_DIR"], exist_ok=True)

        for app_name in settings.INSTALLED_APPS:
            if app_name in self.SKIP_APPS:
                continue

            module = import_module(app_name)
            module_path = Path(module.__file__).parent

            await self.move_apps_static(module_path)
            await self.move_apps_templates(module_path)

        if "starlette_web.contrib.admin" in settings.INSTALLED_APPS:
            await self.move_admin_files()

    async def move_apps_static(self, apps_path: Path):
        static_dir = apps_path / "static"
        if static_dir.is_dir():
            for child in static_dir.iterdir():
                if child.is_file():
                    shutil.copyfile(child, settings.STATIC["ROOT_DIR"] / child.name)
                elif child.is_dir():
                    destination_dir = settings.STATIC["ROOT_DIR"] / child.name
                    if destination_dir.is_dir():
                        shutil.rmtree(destination_dir)
                    shutil.copytree(child, destination_dir)

    async def move_apps_templates(self, apps_path: Path):
        templates_dir = apps_path / "templates"
        if templates_dir.is_dir():
            for child in templates_dir.iterdir():
                if child.is_file():
                    shutil.copyfile(child, settings.TEMPLATES["ROOT_DIR"] / child.name)
                elif child.is_dir():
                    destination_dir = settings.TEMPLATES["ROOT_DIR"] / child.name
                    if destination_dir.is_dir():
                        shutil.rmtree(destination_dir)
                    shutil.copytree(child, destination_dir)

    async def move_admin_files(self):
        # This is special case, since starlette_admin package manages
        # static files in a bit different way
        try:
            module = import_module("starlette_admin")
        except (ImportError, SystemError) as exc:
            raise CommandError(details=str(exc)) from exc

        module_path = Path(module.__file__).parent
        module_static_dir = module_path / "statics"
        module_templates_dir = module_path / "templates"

        if not module_static_dir.is_dir():
            raise CommandError(details="Invalid structure of starlette_admin")
        if not module_templates_dir.is_dir():
            raise CommandError(details="Invalid structure of starlette_admin")

        target_static_dir = settings.STATIC["ROOT_DIR"] / "admin"
        target_templates_dir = settings.TEMPLATES["ROOT_DIR"] / "admin"

        Path(target_static_dir).mkdir(exist_ok=True)
        Path(target_templates_dir).mkdir(exist_ok=True)

        for child in module_static_dir.iterdir():
            if child.is_file():
                shutil.copyfile(child, target_static_dir / child.name)
            elif child.is_dir():
                destination_dir = target_static_dir / child.name
                if destination_dir.is_dir():
                    shutil.rmtree(destination_dir)
                shutil.copytree(child, destination_dir)

        for child in module_templates_dir.iterdir():
            if child.is_file():
                shutil.copyfile(child, target_templates_dir / child.name)
            elif child.is_dir():
                destination_dir = target_templates_dir / child.name
                if destination_dir.is_dir():
                    shutil.rmtree(destination_dir)
                shutil.copytree(child, destination_dir)



import asyncio
import multiprocessing
from unittest.mock import Mock


class BaseMock:
    """Base class for class mocking

    # users class
    >>> class Vehicle:
    >>>    def run(self): ...

    # mock class
    >>> class MockVehicle(BaseMock):
    >>>     target_class = Vehicle
    >>>     def __init__(self):
    >>>         self.run = Mock(return_value=None)  # noqa

    """

    CODE_OK = 0
    target_obj = None

    @property
    def target_class(self):
        raise NotImplementedError

    def get_mocks(self):
        return [attr for attr, val in self.__dict__.items() if callable(val)]

    def mock_init(self, *args, **kwargs):
        ...

    @staticmethod
    def async_return(result):
        f = asyncio.Future()
        f.set_result(result)
        return f


class MockProcess(BaseMock):
    target_class = multiprocessing.context.Process

    def __init__(self):
        self.start = Mock(return_value=None)
        self.terminate = Mock(return_value=None)
        self.__repr__ = Mock(return_value="TestProcess")



import datetime
import uuid

import pytest

from starlette_web.common.http.exceptions import BaseApplicationError
from starlette_web.tests.helpers import await_


def test_constance_get_value(config):
    test_constant_1_value = await_(config.get("TEST_CONSTANT_1"))
    assert test_constant_1_value == 1

    all_values = await_(config.mget(["TEST_CONSTANT_1", "TEST_CONSTANT_2"]))
    assert all_values == {"TEST_CONSTANT_1": 1, "TEST_CONSTANT_2": 2}

    test_constant_datetime = await_(config.get("TEST_CONSTANT_DATETIME"))
    assert type(test_constant_datetime) == datetime.datetime


def test_constance_set_value(config):
    new_value = uuid.uuid4()
    await_(config.set("TEST_CONSTANT_UUID", new_value))
    test_constant_uuid = await_(config.get("TEST_CONSTANT_UUID"))
    assert new_value == test_constant_uuid


def test_constance_errors(config):
    new_value = uuid.uuid4()
    with pytest.raises(BaseApplicationError):
        await_(config.set("TEST_CONSTANT_DATETIME", new_value))

    with pytest.raises(BaseApplicationError):
        await_(config.get("TEST_CONSTANT_NOT_EXISTING_KEY"))



# Copy from https://github.com/vbabiy/djangorestframework-camel-case

import re
from typing import Union
from collections.abc import Container

camelize_re = re.compile(r"[a-z0-9]?_[a-z0-9]")
JSONType = Union[dict, list, str, bool, None]


def _underscore_to_camel(match):
    group = match.group()
    if len(group) == 3:
        return group[0] + group[2].upper()
    else:
        return group[1].upper()


def camelize_key(key):
    return re.sub(camelize_re, _underscore_to_camel, key)


def camelize(data: JSONType, ignore_fields: Container = ()) -> JSONType:
    """
    Recursively swaps snake_case to camelCase in keys of nested dictionaries, leaving values as is.

    >>> from starlette_web.contrib.camel_case import camelize
    >>> camelize({'key_item': 1, 'values_list': [{'object_id': 1, 'object_value': 'new_object_value'}]})  # noqa E501
    [0] {'keyItem': 1, 'valuesList': [{'objectId': 1, 'objectValue': 'new_object_value'}]}
    """
    if isinstance(data, dict):
        new_dict = dict()
        for key, value in data.items():
            if isinstance(key, str) and "_" in key:
                new_key = re.sub(camelize_re, _underscore_to_camel, key)
            else:
                new_key = key

            if key not in ignore_fields and new_key not in ignore_fields:
                new_dict[new_key] = camelize(value, ignore_fields=ignore_fields)
            else:
                new_dict[new_key] = value
        return new_dict

    if _is_iterable(data) and not isinstance(data, str):
        return [camelize(item, ignore_fields=ignore_fields) for item in data]

    return data


def _get_underscoreize_re(no_underscore_before_number=False):
    if no_underscore_before_number:
        pattern = r"([a-z0-9]|[A-Z]?(?=[A-Z](?=[a-z])))([A-Z])"
    else:
        pattern = r"([a-z0-9]|[A-Z]?(?=[A-Z0-9](?=[a-z0-9]|$)))([A-Z]|(?<=[a-z])[0-9](?=[0-9A-Z]|$)|(?<=[A-Z])[0-9](?=[0-9]|$))"  # noqa E501
    return re.compile(pattern)


def _camel_to_underscore(name, no_underscore_before_number=False):
    underscoreize_re = _get_underscoreize_re(
        no_underscore_before_number=no_underscore_before_number
    )
    return underscoreize_re.sub(r"\1_\2", name).lower()


def underscoreize(
    data: JSONType,
    ignore_fields: Container = (),
    no_underscore_before_number: bool = False,
) -> JSONType:
    """
    Recursively swaps camelCase to snake_case in keys of nested dictionaries, leaving values as is.

    >>> from starlette_web.contrib.camel_case import underscoreize
    >>> underscoreize({'keyItem': 1, 'valuesList': [{'objectId': 1, 'objectValue': 'newObjectValue'}]})  # noqa E501
    [0] {'key_item': 1, 'values_list': [{'object_id': 1, 'object_value': 'newObjectValue'}]}
    """
    if isinstance(data, dict):
        new_dict = {}
        for key, value in data.items():
            if isinstance(key, str):
                new_key = _camel_to_underscore(
                    key,
                    no_underscore_before_number=no_underscore_before_number,
                )
            else:
                new_key = key

            if isinstance(value, str):
                new_dict[new_key] = value
            elif (key not in ignore_fields) and (new_key not in ignore_fields):
                new_dict[new_key] = underscoreize(
                    value,
                    ignore_fields=ignore_fields,
                    no_underscore_before_number=no_underscore_before_number,
                )
            else:
                new_dict[new_key] = value

        return new_dict

    if _is_iterable(data) and not isinstance(data, str):
        return [
            underscoreize(
                item,
                ignore_fields=ignore_fields,
                no_underscore_before_number=no_underscore_before_number,
            )
            for item in data
        ]

    return data


def _is_iterable(obj):
    try:
        _ = iter(obj)
        return True
    except TypeError:
        return False



import asyncio
import time
import uuid
from contextlib import contextmanager
from datetime import datetime, timedelta
from hashlib import blake2b
from typing import Tuple, Type
from unittest import mock

from sqlalchemy.ext.asyncio import AsyncSession
from starlette.testclient import TestClient

from starlette_web.common.database import make_session_maker
from starlette_web.contrib.auth.utils import encode_jwt
from starlette_web.contrib.auth.models import User, UserSession
from starlette_web.tests.mocks import BaseMock


class WebTestClient(TestClient):
    db_session: AsyncSession = None

    def login(self, user: User):
        user_session = create_user_session(self.db_session, user)
        jwt, _ = encode_jwt({"user_id": user.id, "session_id": user_session.public_id})
        self.headers["Authorization"] = f"Bearer {jwt}"
        return user_session

    def logout(self):
        self.headers.pop("Authorization", None)


def await_(coroutine):
    """Run coroutine in the current event loop"""

    loop = asyncio.get_event_loop_policy().get_event_loop()
    return loop.run_until_complete(coroutine)


def mock_target_class(mock_class: Type[BaseMock], monkeypatch):
    """Allows to mock any classes (is used as fixture)

    # in conftest.py:
    >>> import pytest
    >>> @pytest.fixture
    >>> def mocked_bicycle(monkeypatch) -> MockBicycle: # noqa
    >>>     yield from mock_target_class(MockBicycle, monkeypatch) # noqa

    # in test.py:
    >>> def test_something(mocked_sender):
    >>>     mocked_bicycle.run.assert_called
    >>>     mocked_bicycle.target_class.__init__.assert_called
    """

    mock_obj = mock_class()

    def init_method(target_obj=None, *args, **kwargs):
        nonlocal mock_obj
        mock_obj.target_obj = target_obj
        mock_obj.mock_init(*args, **kwargs)

    with mock.patch.object(mock_class.target_class, "__init__", autospec=True) as init:
        init.side_effect = init_method
        for mock_method in mock_obj.get_mocks():
            monkeypatch.setattr(
                mock_class.target_class, mock_method, getattr(mock_obj, mock_method)
            )

        yield mock_obj

    del mock_obj


def get_user_data() -> Tuple[str, str]:
    return f"u_{uuid.uuid4().hex[:10]}@test.com", "password"


def get_source_id() -> str:
    """Generate SourceID (ex.: youtube's video-id)"""
    return blake2b(key=bytes(str(time.time()), encoding="utf-8"), digest_size=6).hexdigest()[:11]


@contextmanager
def make_db_session(loop):
    session_maker = make_session_maker()
    async_session = session_maker()
    await_(async_session.__aenter__())
    yield async_session
    await_(async_session.__aexit__(None, None, None))


def create_user(db_session):
    email, password = get_user_data()
    return await_(User.async_create(db_session, db_commit=True, email=email, password=password))


def create_user_session(db_session, user):
    return await_(
        UserSession.async_create(
            db_session,
            db_commit=True,
            user_id=user.id,
            public_id=str(uuid.uuid4()),
            refresh_token="refresh-token",
            is_active=True,
            expired_at=datetime.utcnow() + timedelta(seconds=120),
            created_at=datetime.utcnow(),
            refreshed_at=datetime.utcnow(),
        )
    )



# flake8: noqa

from starlette_web.common.database.columns import EnumTypeColumn
from starlette_web.common.database.model_base import ModelBase
from starlette_web.common.database.model_mixin import ModelMixin, DBModel
from starlette_web.common.database.session_maker import make_session_maker
from starlette_web.common.database.types import ChoiceType



import os
from typing import Type

from starlette_web.common.conf import settings
from starlette_web.common.http.exceptions import ImproperlyConfigured
from starlette_web.common.utils.importing import import_string
from starlette_web.contrib.scheduler.backends.base import BasePeriodicTaskScheduler


def get_periodic_scheduler_backend_class() -> Type[BasePeriodicTaskScheduler]:
    try:
        backend_klass = settings.PERIODIC_JOBS_BACKEND
    except (ImproperlyConfigured, AssertionError):
        backend_klass = None

    if backend_klass is not None:
        return import_string(backend_klass)

    if os.name == "nt":
        return import_string("starlette_web.contrib.scheduler.backends.win32.WindowsTaskScheduler")

    return import_string("starlette_web.contrib.scheduler.backends.posix.CrontabScheduler")



from starlette_web.common.management.base import BaseCommand


class Command(BaseCommand):
    help = "Command to test command line arguments"

    def add_arguments(self, parser):
        parser.add_argument(
            "integers",
            metavar="N",
            type=int,
            nargs="+",
            help="an integer for the accumulator",
        )
        parser.add_argument(
            "--sum",
            dest="accumulate",
            action="store_const",
            const=sum,
            default=max,
            help="sum the integers (default: find the max)",
        )

    async def handle(self, **options):
        print(options["accumulate"](options["integers"]))



import ctypes
import datetime
import json
import logging
import warnings

import croniter
from py_win_task_scheduler import (
    list_folders,
    create_folder,
    list_tasks,
    info as get_task_info,
    create_task,
    delete_task,
)

from starlette_web.common.management.base import CommandError
from starlette_web.contrib.scheduler.backends.base import BasePeriodicTaskScheduler
from starlette_web.contrib.scheduler.app_settings import Win32Settings


logger = logging.getLogger("starlette_web.contrib.scheduler")


class WindowsTaskScheduler(BasePeriodicTaskScheduler):
    settings_class = Win32Settings

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._project_folder_exists = False
        self._current_jobs = []
        self._root_dir = "starlette_web"

    def __enter__(self):
        if self.needs_write:
            if ctypes.windll.shell32.IsUserAnAdmin() == 0:
                warnings.warn(
                    "You may have insufficient privileges to make "
                    "changes to Task Scheduler as non-administrator"
                )

        self._ensure_project_folder_exists()
        self._read_jobs()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False

    def _read_jobs(self):
        project_hash = self._get_project_level_hash()
        location = "\\" + self._root_dir + "\\" + project_hash
        self._current_jobs = list_tasks(location=location)

    def update_jobs(self):
        all_jobs = [self._hash_job(job) for job in self.settings.PERIODIC_JOBS]
        jobs_to_add = set(all_jobs) - set(self._current_jobs)
        jobs_to_delete = set(self._current_jobs) - set(all_jobs)

        project_hash = self._get_project_level_hash()
        location = "\\" + self._root_dir + "\\" + project_hash

        for job_hash in jobs_to_delete:
            deleted = delete_task(job_hash, location=location)
            if not deleted:
                raise CommandError(f"Could not delete task {job_hash}")

            logger.info(f"Removing scheduled task {job_hash}")

        for job_hash in jobs_to_add:
            job = self._get_job_by_hash(job_hash)

            if job[0] == "@reboot":
                schedule_kwargs = dict(
                    trigger_type="OnBoot",
                )
            else:
                # Add 5 seconds to ensure Windows has time to prepare entry in scheduler registry
                now = datetime.datetime.now().astimezone() + datetime.timedelta(seconds=5)
                it = croniter.croniter("* * * * *", now)
                start_dt = datetime.datetime.fromtimestamp(next(it)).astimezone()
                start_time = start_dt.time().strftime("%H:%M")
                start_date = start_dt.date().strftime("%Y-%m-%d")

                schedule_kwargs = dict(
                    trigger_type="Daily",
                    repeat_interval="1 minute",
                    repeat_duration="1 day",
                    start_time=start_time,
                    start_date=start_date,
                )

            created = create_task(
                name=job_hash,
                location=location,
                description=f"Periodic job {job_hash}",
                enabled=True,
                hidden=False,
                user_name=self.settings.USERNAME,
                password=self.settings.PASSWORD,
                run_if_idle=False,
                ac_only=False,
                stop_if_on_batteries=False,
                wake_to_run=True,
                run_if_network=False,
                allow_demand_start=False,
                start_when_available=True,
                force=True,
                restart_every=False,
                execution_time_limit=False,
                force_stop=True,
                delete_after=False,
                multiple_instances=not self.settings.LOCK_JOBS,
                action_type="Execute",
                cmd=self.settings.PYTHON_EXECUTABLE,
                arguments=(
                    f"command.py "
                    f"scheduler "
                    f"run "
                    f"--jobhash={job_hash} "
                    f"--settings={self.settings.SETTINGS_MODULE}"
                ),
                start_in=self.settings.RUN_DIRECTORY,
                **schedule_kwargs,
            )
            if not created:
                raise CommandError(message=f"Could not create task {job}")

            logger.info(f"Created scheduled task {job}")

    def _ensure_project_folder_exists(self):
        if self._project_folder_exists:
            return

        root_folders = list_folders("\\")
        if self._root_dir not in root_folders:
            created = create_folder(self._root_dir, location="\\")
            if not created:
                raise CommandError(
                    details="Root folder for project scheduled tasks "
                    "could not be created with Task Scheduler 2.0"
                )

        project_hash = self._get_project_level_hash()
        projects_folders = list_folders("\\" + self._root_dir)
        if project_hash not in projects_folders:
            created = create_folder(project_hash, location="\\" + self._root_dir)
            if not created:
                raise CommandError(
                    details="Root folder for project scheduled tasks "
                    "could not be created with Task Scheduler 2.0"
                )

        self._project_folder_exists = True

    async def run_job(self, job_hash: str):
        job = self._get_job_by_hash(job_hash)

        # Windows jobs actually run every minute, since there is no easy
        # way to support crontab syntax. Instead, we check it here.
        if job[0] != "@reboot":
            now = datetime.datetime.now().astimezone() - datetime.timedelta(seconds=5)
            next_run_ts = next(croniter.croniter(job[0], now))
            if (
                datetime.datetime.fromtimestamp(next_run_ts).astimezone() - now
            ).total_seconds() > 30:
                return

        return await super().run_job(job_hash)

    def show_jobs(self):
        project_hash = self._get_project_level_hash()
        location = "\\" + self._root_dir + "\\" + project_hash

        for job in self.settings.PERIODIC_JOBS:
            task_name = self._hash_job(job)
            if task_name in self._current_jobs:
                task_info = get_task_info(task_name, location=location)
                logger.info("%s -> %s" % (job, location + task_name))
                logger.info(json.dumps(task_info, indent=2))

    def remove_jobs(self):
        project_hash = self._get_project_level_hash()
        location = "\\" + self._root_dir + "\\" + project_hash

        for job_hash in self._current_jobs:
            deleted = delete_task(name=job_hash, location=location)

            if not deleted:
                raise CommandError(
                    details=f"Task {job_hash} for project scheduled tasks "
                    "could not be deleted with Task Scheduler 2.0"
                )



from starlette_web.common.http.renderers import JSONRenderer
from starlette_web.contrib.camel_case.utils import camelize


class CamelCaseJSONRenderer(JSONRenderer):
    @staticmethod
    def preprocess_content(content):
        return camelize(content)



from typing import Sequence, Any, List, Dict, Type, Optional, AsyncContextManager

from redis import asyncio as aioredis

from starlette_web.common.caches.base import BaseCache, CacheError
from starlette_web.common.http.exceptions import UnexpectedError
from starlette_web.common.utils.encoding import force_str
from starlette_web.common.utils.serializers import BytesSerializer, PickleSerializer
from starlette_web.contrib.redis.redislock import RedisLock


def reraise_exception(func):
    async def wrapped(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except aioredis.RedisError as exc:
            raise CacheError from exc
        except Exception as exc:
            raise UnexpectedError from exc

    return wrapped


class RedisCache(BaseCache):
    redis: aioredis.Redis
    serializer_class: Type[BytesSerializer] = PickleSerializer
    lock_class = RedisLock

    def __init__(self, options: Dict[str, Any]):
        super().__init__(options)
        self.redis = aioredis.Redis(**options)

    @reraise_exception
    async def async_get(self, key: str) -> Any:
        value = await self.redis.get(key)
        return self.serializer.deserialize(value)

    @reraise_exception
    async def async_set(self, key: str, value, timeout: Optional[float] = 120):
        await self.redis.set(
            key,
            self.serializer.serialize(value),
            px=int(timeout * 1000) if timeout is not None else None,
        )

    @reraise_exception
    async def async_delete(self, key: str) -> None:
        await self.redis.delete(key)

    @reraise_exception
    async def async_keys(self, pattern: str) -> List[str]:
        # Using KEYS is not recommended in production with high-load,
        # since it's a redis-blocking operation
        # If you have millions of keys, consider using redis.scan_iter instead
        return [force_str(key) for key in (await self.redis.keys(pattern))]

    @reraise_exception
    async def async_get_many(self, keys: Sequence[str]) -> Dict[str, Any]:
        result = dict()
        key_idx = 0

        # redis.mget returns a simple list
        for value in await self.redis.mget(keys):
            result[keys[key_idx]] = self.serializer.deserialize(value)
            key_idx += 1

        return result

    @reraise_exception
    async def async_has_key(self, key: str) -> bool:
        return bool(await self.redis.exists(key))

    @reraise_exception
    async def async_set_many(self, data: Dict[str, Any], timeout: Optional[float] = 120) -> None:
        """
        Set multiple key-values with timeout.
        Note, that redis does not support setting timeout in MSET,
        which is why commands need to be wrapped with multi-exec block (provided by pipelines)
        """

        if len(data) < 2:
            await super().async_set_many(data, timeout)

        elif timeout is None:
            await self.redis.mset(
                {key: self.serializer.serialize(value) for key, value in data.items()}
            )

        else:
            async with aioredis.client.Pipeline(
                self.redis.connection_pool,
                response_callbacks={},
                transaction=True,
                shard_hint=None,
            ) as pipeline:
                for key, value in data.items():
                    pipeline.execute_command(
                        "SET",
                        key,
                        self.serializer.serialize(value),
                        "PX",
                        int(timeout * 1000),
                    )
                await pipeline.execute(raise_on_error=True)

    @reraise_exception
    async def async_delete_many(self, keys: Sequence[str]) -> None:
        await self.redis.delete(*keys)

    @reraise_exception
    async def async_clear(self) -> None:
        await self.redis.flushdb()

    def lock(
        self,
        name: str,
        timeout: Optional[float] = 20.0,
        blocking_timeout: Optional[float] = None,
        **kwargs,
    ) -> AsyncContextManager:
        # Warning: may hang, if sleep (default 0.1) is less than timeout/blocking_timeout
        # May be connected to https://github.com/redis/redis-py/issues/2579
        kwargs.pop("sleep", 0)

        return self.redis.lock(
            name,
            timeout=timeout,
            blocking_timeout=blocking_timeout,
            lock_class=self.lock_class,
            sleep=0,
            **kwargs,
        )



import os
import shutil
from pathlib import Path

from starlette_web.common.conf.app_manager import app_manager
from starlette_web.common.management.base import BaseCommand, CommandError, CommandParser


class Command(BaseCommand):
    help = "Initialize application directory"

    def add_arguments(self, parser: CommandParser):
        parser.add_argument("app_name", type=str)

    async def handle(self, **options):
        current_dir = os.getcwd()
        app_name = options["app_name"]

        cwd = Path(current_dir)
        if not (cwd / "command.py").is_file():
            raise CommandError(
                details="This command may only be run in project root directory."
            )

        app_dir = cwd / app_name
        if app_dir.is_file() or app_dir.is_symlink():
            raise CommandError(
                details=(
                    f"Cannot create application directory {app_dir}. "
                    "A file/link with such name exists in the current directory."
                )
            )

        if app_dir.is_dir():
            raise CommandError(
                details=f"Directory {app_dir} already exists. Exiting."
            )

        app_manager.register_apps()
        if app_name in app_manager.app_names:
            existing_app = app_manager.app_names[app_name]
            raise CommandError(
                details=f"Application with name {app_name} already installed in "
                        f"settings.INSTALLED_APPS ({existing_app})."
            )

        app_dir.mkdir()
        defaults_dir = Path(__file__).parent / "_app_defaults"

        for source_file in defaults_dir.iterdir():
            if source_file.is_file() and source_file.name.endswith(".py"):
                target_file = app_dir / source_file.name
                shutil.copy(source_file, target_file)

        with open(app_dir / "apps.py", "rt") as file:
            lines = []
            for line in file:
                lines.append(line.replace("__APPNAME", app_name))

        with open(app_dir / "apps.py", "wt") as file:
            file.writelines(lines)



import base64
import hashlib
import math
from pathlib import Path
import pickle
import re
import tempfile
import time
from typing import AsyncContextManager, Optional, Sequence, Dict, Any, List, BinaryIO, Type

from anyio.lowlevel import cancel_shielded_checkpoint

from starlette_web.common.conf import settings
from starlette_web.common.caches.base import BaseCache, CacheError
from starlette_web.common.files.filelock import FileLock
from starlette_web.common.http.exceptions import ImproperlyConfigured
from starlette_web.common.utils.regex import redis_pattern_to_re_pattern
from starlette_web.common.utils.serializers import (
    BytesSerializer,
    PickleSerializer,
    DeserializeError,
)


class FileCache(BaseCache):
    # For test purposes, not recommended for production
    lock_class = FileLock
    _manager_lock_blocking_timeout = 10
    _manager_lock_timeout = 1.0
    serializer_class: Type[BytesSerializer] = PickleSerializer
    timestamp_bom = b"ND16C7Bh9Xd"

    def __init__(self, options: Dict[str, Any]):
        super().__init__(options)
        self.base_dir = options.get("CACHE_DIR")
        if self.base_dir is None or not Path(self.base_dir).is_dir():
            raise ImproperlyConfigured(
                details="Invalid CACHE_DIR value for FileCache"
            )

        self.serializer = self.serializer_class()
        if not self.serializer.serializes_to_bytes():
            raise ImproperlyConfigured(
                details="serializer_class must be instance of BytesSerializer"
            )

    async def async_get(self, key: str) -> Any:
        async with self._get_manager_lock():
            return self._sync_get(key)

    async def async_get_many(self, keys: Sequence[str]) -> Dict[str, Any]:
        async with self._get_manager_lock():
            results = {}
            for key in keys:
                await cancel_shielded_checkpoint()
                results[key] = self._sync_get(key)
            return results

    def _sync_get(self, key: str) -> Any:
        path = Path(self.base_dir) / self._key_name(key)
        if path.is_file():
            with open(path, "rb") as file:
                if self._sync_get_deadline(file) < time.time():
                    return None
                content = self.serializer.deserialize(file.read())
                return content["data"]
        return None

    def _sync_get_deadline(self, _file: BinaryIO) -> float:
        try:
            deadline = _file.read(32)
            if deadline[:11] == self.timestamp_bom:
                # pickle.dumps of float is always 21 bytes
                return pickle.loads(deadline[-21:])
            else:
                _file.seek(0)
                return math.inf

        except DeserializeError:
            _file.seek(0)
            return math.inf

    async def async_set(self, key: str, value: Any, timeout: Optional[float] = 120) -> None:
        async with self._get_manager_lock():
            self._sync_set(key, value, timeout)

    async def async_set_many(self, data: Dict[str, Any], timeout: Optional[float] = 120) -> None:
        async with self._get_manager_lock():
            for key, value in data.items():
                await cancel_shielded_checkpoint()
                self._sync_set(key, value, timeout)

    def _sync_set(self, key: str, value: Any, timeout: Optional[float] = 120) -> None:
        path = Path(self.base_dir) / self._key_name(key)
        with open(path, "wb") as file:
            content = {"data": value}
            if timeout is not None:
                deadline = time.time() + timeout
                # pickle.dumps of float is always 21 bytes
                file.write(self.timestamp_bom + pickle.dumps(deadline, protocol=4))
            file.write(self.serializer.serialize(content))

    async def async_delete(self, key: str) -> None:
        async with self._get_manager_lock():
            self._sync_delete(key)

    async def async_delete_many(self, keys: Sequence[str]) -> None:
        async with self._get_manager_lock():
            for key in keys:
                self._sync_delete(key)

    def _sync_delete(self, key) -> None:
        path = Path(self.base_dir) / self._key_name(key)
        if path.is_file():
            path.unlink()

    async def async_clear(self) -> None:
        async with self._get_manager_lock():
            _manager_lock_name = self._get_manager_lock_name()
            for file in Path(self.base_dir).iterdir():
                if str(file) == _manager_lock_name:
                    continue
                file.unlink()

    async def async_keys(self, pattern: str) -> List[str]:
        try:
            re_pattern = redis_pattern_to_re_pattern(pattern)
        except re.error as exc:
            raise CacheError(details=str(exc)) from exc

        keys = []
        async with self._get_manager_lock():
            _manager_lock_name = self._get_manager_lock_name()
            for file in Path(self.base_dir).iterdir():
                await cancel_shielded_checkpoint()
                if str(file) == _manager_lock_name:
                    continue
                key = base64.b32decode(file.name.replace("8", "=").encode()).decode()
                # Check for expiration
                with open(str(file), "rb") as file:
                    deadline = self._sync_get_deadline(file)
                if deadline >= time.time() and re.fullmatch(re_pattern, key):
                    keys.append(key)
            return keys

    async def async_has_key(self, key: str) -> bool:
        return (await self.async_get(key)) is not None

    def _get_manager_lock(self):
        return self.lock_class(
            self._get_manager_lock_name(),
            blocking_timeout=self._manager_lock_blocking_timeout,
            timeout=self._manager_lock_timeout,
        )

    def _get_manager_lock_name(self) -> str:
        return str(Path(tempfile.gettempdir()) / (self._get_project_hash() + '_cache.lock'))

    @staticmethod
    def _key_name(key: str) -> str:
        # 151-155 base256 => 248 base32
        # 156-160 base256 => 256 base32
        # Linux supports only 255 bytes for filename,
        # as for Windows, lengthy paths also not recommended
        encoded_key = key.encode("utf-8")
        if len(encoded_key) >= 156:
            raise CacheError(details="Key length must not exceed 155 bytes.")

        return base64.b32encode(encoded_key).decode().replace("=", "8")

    def lock(
        self,
        name: str,
        timeout: Optional[float] = 20.0,
        blocking_timeout: Optional[float] = None,
        **kwargs,
    ) -> AsyncContextManager:
        return self.lock_class(
            name=name,
            timeout=timeout,
            blocking_timeout=blocking_timeout,
        )

    def _get_project_hash(self):
        return hashlib.md5(str(settings.SECRET_KEY).encode('utf-8')).hexdigest()



# flake8: noqa

from pathlib import Path

from starlette.config import Config
from starlette.datastructures import Secret

PROJECT_ROOT_DIR = Path(__file__).parent.parent.parent

config = Config(PROJECT_ROOT_DIR / ".env")

APP_DEBUG = config("APP_DEBUG", cast=bool, default=False)
SECRET_KEY = config("SECRET_KEY", cast=Secret, default="project-secret")

TEST_MODE = True

INSTALLED_APPS = [
    "starlette_web.contrib.staticfiles",
    "starlette_web.contrib.apispec",
    "starlette_web.contrib.auth",
    "starlette_web.contrib.admin",
    "starlette_web.contrib.constance",
    "starlette_web.contrib.constance.backends.database",
    "starlette_web.contrib.scheduler",
    "starlette_web.tests",
]

DB_ECHO = config("DB_ECHO", cast=bool, default=False)
DB_NAME = config("DB_NAME_TEST", default="web_project_test")

# TODO: refactor database-specific settings with respect to sqlalchemy.create_engine / asyncpg
DATABASE = {
    "driver": "postgresql+asyncpg",
    "host": config("DB_HOST", default="localhost"),
    "port": config("DB_PORT", cast=int, default=5432),
    "username": config("DB_USERNAME", default="starlette-web"),
    "password": config("DB_PASSWORD", cast=Secret, default=None),
    "database": DB_NAME,
    "pool_min_size": config("DB_POOL_MIN_SIZE", cast=int, default=1),
    "pool_max_size": config("DB_POOL_MAX_SIZE", cast=int, default=16),
    "ssl": config("DB_SSL", default=None),
    "use_connection_for_request": config("DB_USE_CONNECTION_FOR_REQUEST", cast=bool, default=True),
    "retry_limit": config("DB_RETRY_LIMIT", cast=int, default=1),
    "retry_interval": config("DB_RETRY_INTERVAL", cast=int, default=1),
}

DATABASE_DSN = config(
    "DB_DSN",
    cast=str,
    default="{driver}://{username}:{password}@{host}:{port}/{database}".format(**DATABASE),
)

ROUTES = "starlette_web.tests.routes.routes"

FILECACHE_DIR = PROJECT_ROOT_DIR / "filecache"
FILECACHE_DIR.mkdir(exist_ok=True)

FILESTORAGE_DIR = PROJECT_ROOT_DIR / "filestorage"
FILESTORAGE_DIR.mkdir(exist_ok=True)

CACHES = {
    "default": {
        "BACKEND": "starlette_web.contrib.redis.RedisCache",
        "OPTIONS": {
            "host": config("REDIS_HOST", default="localhost"),
            "port": config("REDIS_PORT", default=6379),
            "db": config("REDIS_DB", default=0),
        },
    },
    "locmem": {
        "BACKEND": "starlette_web.common.caches.local_memory.LocalMemoryCache",
        "OPTIONS": {
            "name": "locmem",
        },
    },
    "files": {
        "BACKEND": "starlette_web.common.files.cache.FileCache",
        "OPTIONS": {
            "CACHE_DIR": FILECACHE_DIR,
        },
    },
}

CHANNEL_LAYERS = {
    "redispubsub": {
        "BACKEND": "starlette_web.contrib.redis.channel_layers.RedisPubSubChannelLayer",
        "OPTIONS": {
            "host": config("REDIS_HOST", default="localhost"),
            "port": config("REDIS_PORT", default=6379),
            "db": config("REDIS_DB", default=0),
        },
    },
}

SITE_URL = config("SITE_URL", default="") or "https://web.project.com"

LOG_LEVEL = config("LOG_LEVEL", default="INFO")
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {
            "format": "[%(asctime)s] %(levelname)s [%(name)s:%(lineno)s] %(message)s",
            "datefmt": "%d.%m.%Y %H:%M:%S",
        },
    },
    "handlers": {
        "console": {"class": "logging.StreamHandler", "formatter": "standard"},
    },
    "loggers": {
        "uvicorn": {"handlers": ["console"], "level": LOG_LEVEL, "propagate": False},
        "starlette_web": {"handlers": ["console"], "level": LOG_LEVEL, "propagate": False},
        "common": {"handlers": ["console"], "level": LOG_LEVEL, "propagate": False},
        "app": {"handlers": ["console"], "level": LOG_LEVEL, "propagate": False},
    },
}

TEMPLATES = {
    "ROOT_DIR": PROJECT_ROOT_DIR / "templates",
    "AUTOESCAPE": False,
    "AUTORELOAD": False,
}

STATIC = {
    "ROOT_DIR": PROJECT_ROOT_DIR / "static",
    "URL": "/static/",
}
STATIC["ROOT_DIR"].mkdir(exist_ok=True)

MEDIA = {
    "ROOT_DIR": PROJECT_ROOT_DIR / "media",
    "URL": "/media/",
}
MEDIA["ROOT_DIR"].mkdir(exist_ok=True)

# TODO: Think about managing permissions
APISPEC = {
    "CONFIG": dict(
        title="Project documentation",
        version="0.0.1",
        openapi_version="3.0.2",
        info=dict(description="My custom project."),
    ),
    "CONVERT_TO_CAMEL_CASE": False,
}



import os
import sys
from typing import Any

from starlette_web.common.http.exceptions import ImproperlyConfigured


class Settings:
    def __init__(self, project_settings):
        self.LOCK_JOBS = self._getattr(project_settings, "PERIODIC_JOBS_LOCK", True)
        self.PERIODIC_JOBS = self._getattr(project_settings, "PERIODIC_JOBS_LIST", [])
        self.PYTHON_EXECUTABLE = self._getattr(
            project_settings, "PERIODIC_JOBS_PYTHON_EXECUTABLE", sys.executable
        )
        self.RUN_DIRECTORY = self._getattr(
            project_settings, "PERIODIC_JOBS_RUN_DIRECTORY", os.getcwd()
        )
        self.SETTINGS_MODULE = self._getattr(project_settings, "STARLETTE_SETTINGS_MODULE", None)
        self.BLOCKING_TIMEOUT = 5

    @staticmethod
    def _getattr(settings: Any, key: str, default: Any) -> Any:
        try:
            return getattr(settings, key)
        except (ImproperlyConfigured, AttributeError):
            return default


class PosixSettings(Settings):
    def __init__(self, project_settings):
        super().__init__(project_settings)
        self.COMMAND_PREFIX = self._getattr(
            project_settings,
            "PERIODIC_JOBS_COMMAND_PREFIX",
            f"cd {self.RUN_DIRECTORY} &&",
        )
        self.COMMAND_SUFFIX = self._getattr(
            project_settings,
            "PERIODIC_JOBS_COMMAND_SUFFIX",
            "",
        )
        self.CRONTAB_EXECUTABLE = self._getattr(
            project_settings,
            "PERIODIC_JOBS_CRONTAB_EXECUTABLE",
            "/usr/bin/crontab",
        )
        self.CRONTAB_LINE_PATTERN = self._getattr(
            project_settings,
            "PERIODIC_JOBS_CRONTAB_LINE_PATTERN",
            "%(time)s %(command)s # %(comment)s",
        )


class Win32Settings(Settings):
    def __init__(self, project_settings):
        super().__init__(project_settings)
        self.USERNAME = self._getattr(
            project_settings,
            "PERIODIC_JOBS_USERNAME",
            "System",
        )
        self.PASSWORD = self._getattr(
            project_settings,
            "PERIODIC_JOBS_PASSWORD",
            None,
        )



import uuid

from sqlalchemy.ext.asyncio import AsyncSession

from starlette_web.contrib.auth.models import User
from starlette_web.tests.helpers import await_


def test_async_create(dbs: AsyncSession):
    email = str(uuid.uuid4()).replace("-", "") + "@test.com"
    password = User.make_password(str(uuid.uuid4()))

    user = await_(User.async_get(db_session=dbs, email=email, password=password))
    assert user is None

    user = await_(User.async_create(db_session=dbs, email=email, password=password))
    assert user is not None
    assert type(user) == User
    assert user.id is not None

    user = await_(User.async_get(db_session=dbs, email=email, password=password))
    assert user is not None


def test_async_update(dbs: AsyncSession):
    email = str(uuid.uuid4()).replace("-", "") + "@test.com"
    password = User.make_password(str(uuid.uuid4()))
    user = await_(User.async_create(db_session=dbs, email=email, password=password))

    email = str(uuid.uuid4()).replace("-", "") + "@test.com"
    password = User.make_password(str(uuid.uuid4()))
    await_(
        User.async_update(
            db_session=dbs,
            filter_kwargs={"id": user.id},
            update_data={"email": email, "password": password},
        )
    )

    user = await_(User.async_get(db_session=dbs, id=user.id))
    assert user.email == email
    assert user.password == password


def test_async_filter(dbs: AsyncSession):
    email = str(uuid.uuid4()).replace("-", "") + "@test.com"
    password = User.make_password(str(uuid.uuid4()))
    _ = await_(User.async_create(db_session=dbs, email=email, password=password))

    email = str(uuid.uuid4()).replace("-", "") + "@test.com"
    password = User.make_password(str(uuid.uuid4()))
    _ = await_(User.async_create(db_session=dbs, email=email, password=password))

    users = (await_(User.async_filter(db_session=dbs))).all()
    assert len(users) >= 2
    assert type(users[0]) == User


def test_async_create_or_update(dbs: AsyncSession):
    email = str(uuid.uuid4()).replace("-", "") + "@test.com"
    password = User.make_password(str(uuid.uuid4()))

    user = await_(User.async_get(db_session=dbs, email=email, password=password))
    assert user is None

    user = await_(
        User.async_create_or_update(
            db_session=dbs,
            filter_kwargs={"email": email, "password": password},
            update_data={},
        )
    )
    user_id = user.id
    assert user is not None
    assert type(user) == User
    assert user.email == email
    assert user.password == password
    assert user.id is not None

    user = await_(
        User.async_create_or_update(
            db_session=dbs,
            filter_kwargs={"email": email, "password": password},
            update_data={"is_active": True, "is_superuser": True},
        )
    )
    assert user is not None
    assert type(user) == User
    assert user.id == user_id
    assert user.is_superuser

    new_email = str(uuid.uuid4()).replace("-", "") + "@test.com"
    new_password = User.make_password(str(uuid.uuid4()))
    user = await_(
        User.async_create_or_update(
            db_session=dbs,
            filter_kwargs={"email": email, "password": password},
            update_data={"email": new_email, "password": new_password},
        )
    )
    assert user.id == user_id
    assert user.email == new_email
    assert user.password == new_password


def test_async_get_or_create(dbs: AsyncSession):
    email = str(uuid.uuid4()).replace("-", "") + "@test.com"
    password = User.make_password(str(uuid.uuid4()))
    user = await_(
        User.async_get_or_create(
            db_session=dbs, filter_kwargs={"email": email, "password": password}
        )
    )
    user_id = user.id

    assert user is not None
    assert type(user) == User
    assert user.email == email
    assert user.password == password
    assert not user.is_superuser
    assert user_id is not None

    user = await_(
        User.async_get_or_create(
            db_session=dbs, filter_kwargs={"email": email, "password": password}
        )
    )
    assert user.id == user_id

    email = str(uuid.uuid4()).replace("-", "") + "@test.com"
    password = User.make_password(str(uuid.uuid4()))
    user = await_(
        User.async_get_or_create(
            db_session=dbs,
            filter_kwargs={"email": email, "password": password},
            extra_data={"is_superuser": True},
        )
    )
    assert user is not None
    assert type(user) == User
    assert user.is_superuser
    assert user.id is not None



from importlib import import_module

from starlette_web.common.conf import settings
from starlette_web.common.conf.base_app_config import BaseAppConfig


class AppConfig(BaseAppConfig):
    app_name = "admin"
    app_requirements = [
        "starlette_web.contrib.auth",
        "starlette_web.contrib.staticfiles",
    ]

    def initialize(self):
        for installed_app in settings.INSTALLED_APPS:
            try:
                import_module(installed_app + ".admin")
            except (SystemError, ImportError):
                pass
